import datetime
from datetime import timedelta

from colorclass.toggles import disable_all_colors
from sqlalchemy import desc
from sqlalchemy.orm.exc import NoResultFound

from flexget import options
from flexget.event import event
from flexget.manager import Session
from flexget.terminal import TerminalTable, TerminalTableError, colorize, console, table_parser

from . import db


def do_cli(manager, options):
    if options.table_type == 'porcelain':
        disable_all_colors()
    if options.task:
        do_cli_task(manager, options)
    else:
        do_cli_summary(manager, options)


def do_cli_task(manager, options):
    header = ['Start', 'Duration', 'Produced', 'Accepted', 'Rejected', 'Failed', 'Abort Reason']
    table_data = [header]
    with Session() as session:
        try:
            task = session.query(db.StatusTask).filter(db.StatusTask.name == options.task).one()
        except NoResultFound:
            console('Task name `%s` does not exists or does not have any records' % options.task)
            return
        else:
            query = task.executions.order_by(desc(db.TaskExecution.start))[: options.limit]
            for ex in reversed(query):
                start = ex.start.strftime('%Y-%m-%d %H:%M')
                start = colorize('green', start) if ex.succeeded else colorize('red', start)

                if ex.end is not None and ex.start is not None:
                    delta = ex.end - ex.start
                    duration = '%1.fs' % delta.total_seconds()
                else:
                    duration = '?'

                table_data.append(
                    [
                        start,
                        duration,
                        ex.produced,
                        ex.accepted,
                        ex.rejected,
                        ex.failed,
                        ex.abort_reason if ex.abort_reason is not None else '',
                    ]
                )

    try:
        table = TerminalTable(options.table_type, table_data)
        console(table.output)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))


def do_cli_summary(manager, options):
    header = [
        'Task',
        'Last execution',
        'Last success',
        'Produced',
        'Accepted',
        'Rejected',
        'Failed',
        'Duration',
    ]
    table_data = [header]

    with Session() as session:
        for task in session.query(db.StatusTask).all():
            ok = (
                session.query(db.TaskExecution)
                .filter(db.TaskExecution.task_id == task.id)
                .filter(db.TaskExecution.succeeded == True)
                .filter(db.TaskExecution.produced > 0)
                .order_by(db.TaskExecution.start.desc())
                .first()
            )

            if ok is None:
                duration = None
                last_success = '-'
            else:
                duration = ok.end - ok.start
                last_success = ok.start.strftime('%Y-%m-%d %H:%M')

                age = datetime.datetime.utcnow() - ok.start
                if age > timedelta(days=7):
                    last_success = colorize('red', last_success)
                elif age < timedelta(minutes=10):
                    last_success = colorize('green', last_success)
            # Fix weird issue that a task registers StatusTask but without an execution. GH #2022
            last_exec = (
                task.last_execution_time.strftime('%Y-%m-%d %H:%M')
                if task.last_execution_time
                else '-'
            )

            table_data.append(
                [
                    task.name,
                    last_exec,
                    last_success,
                    ok.produced if ok is not None else '-',
                    ok.accepted if ok is not None else '-',
                    ok.rejected if ok is not None else '-',
                    ok.failed if ok is not None else '-',
                    '%1.fs' % duration.total_seconds() if duration is not None else '-',
                ]
            )

    table = TerminalTable(options.table_type, table_data)
    try:
        console(table.output)
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))


@event('options.register')
def register_parser_arguments():
    parser = options.register_command(
        'status', do_cli, help='View task health status', parents=[table_parser]
    )
    parser.add_argument(
        '--task', action='store', metavar='TASK', help='Limit to results in specified %(metavar)s'
    )
    parser.add_argument(
        '--limit',
        action='store',
        type=int,
        metavar='NUM',
        default=50,
        help='Limit to %(metavar)s results',
    )



import hashlib
import http.client
import os
import posixpath
import xml.sax
from datetime import datetime
from urllib.parse import urlparse, urlsplit

import dateutil.parser
import feedparser
from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.pathscrub import pathscrub
from flexget.utils.tools import decode_html

logger = logger.bind(name='rss')
feedparser.registerDateHandler(lambda date_string: dateutil.parser.parse(date_string).timetuple())


def fp_field_name(name):
    """Translates literal field name to the sanitized one feedparser will use."""
    return name.replace(':', '_').lower()


class InputRSS:
    """
    Parses RSS feed.

    Hazzlefree configuration for public rss feeds::

      rss: <url>

    Configuration with basic http authentication::

      rss:
        url: <url>
        username: <name>
        password: <password>

    Advanced usages:

    You may wish to clean up the entry by stripping out all non-ascii characters.
    This can be done by setting ascii value to yes.

    Example::

      rss:
        url: <url>
        ascii: yes

    In case RSS-feed uses some nonstandard field for urls and automatic detection fails
    you can configure plugin to use url from any feedparser entry attribute.

    Example::

      rss:
        url: <url>
        link: guid

    If you want to keep information in another rss field attached to the flexget entry,
    you can use the other_fields option.

    Example::

      rss:
        url: <url>
        other_fields: [date]

    You can disable few possibly annoying warnings by setting silent value to
    yes on feeds where there are frequently invalid items.

    Example::

      rss:
        url: <url>
        silent: yes

    You can group all the links of an item, to make the download plugin tolerant
    to broken urls: it will try to download each url until one works.
    Links are enclosures plus item fields given by the link value, in that order.
    The value to set is "group_links".

    Example::

      rss:
        url: <url>
        group_links: yes
    """

    schema = {
        'type': ['string', 'object'],
        # Simple form, just url or file
        'anyOf': [{'format': 'url'}, {'format': 'file'}],
        # Advanced form, with options
        'properties': {
            'url': {'type': 'string', 'anyOf': [{'format': 'url'}, {'format': 'file'}]},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'title': {'type': 'string'},
            'link': one_or_more({'type': 'string'}),
            'silent': {'type': 'boolean', 'default': False},
            'ascii': {'type': 'boolean', 'default': False},
            'escape': {'type': 'boolean', 'default': False},
            'filename': {'type': 'boolean'},
            'group_links': {'type': 'boolean', 'default': False},
            'all_entries': {'type': 'boolean', 'default': True},
            'other_fields': {
                'type': 'array',
                'items': {
                    # Items can be a string, or a dict with a string value
                    'type': ['string', 'object'],
                    'additionalProperties': {'type': 'string'},
                },
            },
        },
        'required': ['url'],
        'additionalProperties': False,
    }

    def build_config(self, config):
        """Set default values to config"""
        if isinstance(config, str):
            config = {'url': config}
        else:
            # Make a copy so that original config is not modified
            config = dict(config)
        # set the default link value to 'auto'
        config.setdefault('link', 'auto')
        # Convert any field names from the config to format feedparser will use for 'link', 'title' and 'other_fields'
        if config['link'] != 'auto':
            if not isinstance(config['link'], list):
                config['link'] = [config['link']]
            config['link'] = list(map(fp_field_name, config['link']))
        config.setdefault('title', 'title')
        config['title'] = fp_field_name(config['title'])
        if config.get('other_fields'):
            other_fields = []
            for item in config['other_fields']:
                if isinstance(item, str):
                    key, val = item, item
                else:
                    key, val = list(item.items())[0]
                other_fields.append({fp_field_name(key): val.lower()})
            config['other_fields'] = other_fields
        # set default value for group_links as deactivated
        config.setdefault('group_links', False)
        # set default for all_entries
        config.setdefault('all_entries', True)
        return config

    def process_invalid_content(self, task, data, url):
        """If feedparser reports error, save the received data and log error."""

        if data is None:
            logger.critical('Received empty page - no content')
            return
        else:
            data = bytes(data)  # ahem, dunno about this?

        ext = 'xml'
        if b'<html>' in data.lower():
            logger.critical('Received content is HTML page, not an RSS feed')
            ext = 'html'
        if b'login' in data.lower() or b'username' in data.lower():
            logger.critical('Received content looks a bit like login page')
        if b'error' in data.lower():
            logger.critical('Received content looks a bit like error page')
        received = os.path.join(task.manager.config_base, 'received')
        if not os.path.isdir(received):
            os.mkdir(received)
        filename = task.name
        sourcename = urlparse(url).netloc
        if sourcename:
            filename += '-' + sourcename
        filename = pathscrub(filename, filename=True)
        filepath = os.path.join(received, '%s.%s' % (filename, ext))
        with open(filepath, 'wb') as f:
            f.write(data)
        logger.critical('I have saved the invalid content to {} for you to view', filepath)

    def escape_content(self, content):
        valid_escapes = (b'&quot;', b'&apos;', b'&lt;', b'&gt;', b'&amp;')
        future_result = []
        in_cdata_block = False

        for idx, char in enumerate(bytes(content)):

            char = bytes([char])
            if not in_cdata_block and char == b'&':
                if not content[idx : idx + 7].startswith(valid_escapes):
                    char = b'&amp;'
            elif not in_cdata_block and char == b'<' and content[idx : idx + 9] == b'<![CDATA[':
                in_cdata_block = True
            elif in_cdata_block and char == b']' and content[idx - 1 : idx + 2] == b']]>':
                in_cdata_block = False
            future_result.append(char)
        return b''.join(future_result)

    def add_enclosure_info(self, entry, enclosure, filename=True, multiple=False):
        """Stores information from an rss enclosure into an Entry."""
        entry['url'] = enclosure['href']
        # get optional meta-data
        if 'length' in enclosure:
            try:
                entry['size'] = int(enclosure['length'])
            except ValueError:
                entry['size'] = 0
        if 'type' in enclosure:
            entry['type'] = enclosure['type']
        # TODO: better and perhaps join/in download plugin?
        # Parse filename from enclosure url
        basename = posixpath.basename(urlsplit(entry['url']).path)
        # If enclosure has size OR there are multiple enclosures use filename from url
        if (entry.get('size') or multiple and basename) and filename:
            entry['filename'] = basename
            logger.trace('filename `{}` from enclosure', entry['filename'])

    @cached('rss')
    @plugin.internet(logger)
    def on_task_input(self, task, config):
        config = self.build_config(config)

        logger.debug('Requesting task `{}` url `{}`', task.name, config['url'])

        # Used to identify which etag/modified to use
        url_hash = hashlib.md5(config['url'].encode('utf-8')).hexdigest()

        # set etag and last modified headers if config has not changed since
        # last run and if caching wasn't disabled with --no-cache argument.
        all_entries = (
            config['all_entries']
            or task.config_modified
            or task.options.nocache
            or task.options.retry
        )
        headers = task.requests.headers
        if not all_entries:
            etag = task.simple_persistence.get('%s_etag' % url_hash, None)
            if etag:
                logger.debug('Sending etag {} for task {}', etag, task.name)
                headers['If-None-Match'] = etag
            modified = task.simple_persistence.get('%s_modified' % url_hash, None)
            if modified:
                if not isinstance(modified, str):
                    logger.debug('Invalid date was stored for last modified time.')
                else:
                    headers['If-Modified-Since'] = modified
                    logger.debug(
                        'Sending last-modified {} for task {}',
                        headers['If-Modified-Since'],
                        task.name,
                    )

        # Get the feed content
        if config['url'].startswith(('http', 'https', 'ftp', 'file')):
            # Get feed using requests library
            auth = None
            if 'username' in config and 'password' in config:
                auth = (config['username'], config['password'])
            try:
                # Use the raw response so feedparser can read the headers and status values
                response = task.requests.get(
                    config['url'], timeout=60, headers=headers, raise_status=False, auth=auth
                )
                content = response.content
            except RequestException as e:
                raise plugin.PluginError(
                    'Unable to download the RSS for task %s (%s): %s'
                    % (task.name, config['url'], e)
                )
            if config.get('ascii'):
                # convert content to ascii (cleanup), can also help with parsing problems on malformed feeds
                content = response.text.encode('ascii', 'ignore')

            # status checks
            status = response.status_code
            if status == 304:
                logger.verbose(
                    '{} hasn\'t changed since last run. Not creating entries.', config['url']
                )
                # Let details plugin know that it is ok if this feed doesn't produce any entries
                task.no_entries_ok = True
                return []
            elif status == 401:
                raise plugin.PluginError(
                    'Authentication needed for task %s (%s): %s'
                    % (task.name, config['url'], response.headers['www-authenticate']),
                    logger,
                )
            elif status == 404:
                raise plugin.PluginError(
                    'RSS Feed %s (%s) not found' % (task.name, config['url']), logger
                )
            elif status == 500:
                raise plugin.PluginError(
                    'Internal server exception on task %s (%s)' % (task.name, config['url']),
                    logger,
                )
            elif status != 200:
                raise plugin.PluginError(
                    'HTTP error %s received from %s' % (status, config['url']), logger
                )

            # update etag and last modified
            if not config['all_entries']:
                etag = response.headers.get('etag')
                if etag:
                    task.simple_persistence['%s_etag' % url_hash] = etag
                    logger.debug('etag {} saved for task {}', etag, task.name)
                if response.headers.get('last-modified'):
                    modified = response.headers['last-modified']
                    task.simple_persistence['%s_modified' % url_hash] = modified
                    logger.debug('last modified {} saved for task {}', modified, task.name)
        else:
            # This is a file, open it
            with open(config['url'], 'rb') as f:
                content = f.read()
            if config.get('ascii'):
                # Just assuming utf-8 file in this case
                content = content.decode('utf-8', 'ignore').encode('ascii', 'ignore')

        if not content:
            logger.error('No data recieved for rss feed.')
            return []
        if config.get('escape'):
            logger.debug("Trying to escape unescaped in RSS")
            content = self.escape_content(content)
        try:
            rss = feedparser.parse(content)
        except LookupError as e:
            raise plugin.PluginError('Unable to parse the RSS (from %s): %s' % (config['url'], e))

        # check for bozo
        ex = rss.get('bozo_exception', False)
        if ex or rss.get('bozo'):
            if rss.entries:
                msg = (
                    'Bozo error %s while parsing feed, but entries were produced, ignoring the error.'
                    % type(ex)
                )
                if config.get('silent', False):
                    logger.debug(msg)
                else:
                    logger.verbose(msg)
            else:
                if isinstance(ex, feedparser.NonXMLContentType):
                    # see: http://www.feedparser.org/docs/character-encoding.html#advanced.encoding.nonxml
                    logger.debug('ignoring feedparser.NonXMLContentType')
                elif isinstance(ex, feedparser.CharacterEncodingOverride):
                    # see: ticket 88
                    logger.debug('ignoring feedparser.CharacterEncodingOverride')
                elif isinstance(ex, UnicodeEncodeError):
                    raise plugin.PluginError('Feed has UnicodeEncodeError while parsing...')
                elif isinstance(
                    ex, (xml.sax._exceptions.SAXParseException, xml.sax._exceptions.SAXException)
                ):
                    # save invalid data for review, this is a bit ugly but users seem to really confused when
                    # html pages (login pages) are received
                    self.process_invalid_content(task, content, config['url'])
                    if task.options.debug:
                        logger.error('bozo error parsing rss: {}', ex)
                    raise plugin.PluginError(
                        'Received invalid RSS content from task %s (%s)'
                        % (task.name, config['url'])
                    )
                elif isinstance(ex, http.client.BadStatusLine) or isinstance(ex, OSError):
                    raise ex  # let the @internet decorator handle
                else:
                    # all other bozo errors
                    self.process_invalid_content(task, content, config['url'])
                    raise plugin.PluginError(
                        'Unhandled bozo_exception. Type: %s (task: %s)'
                        % (ex.__class__.__name__, task.name),
                        logger,
                    )

        logger.debug('encoding {}', rss.encoding)

        last_entry_id = ''
        if not all_entries:
            # Test to make sure entries are in descending order
            if (
                rss.entries
                and rss.entries[0].get('published_parsed')
                and rss.entries[-1].get('published_parsed')
            ):
                if rss.entries[0]['published_parsed'] < rss.entries[-1]['published_parsed']:
                    # Sort them if they are not
                    rss.entries.sort(key=lambda x: x['published_parsed'], reverse=True)
            last_entry_id = task.simple_persistence.get('%s_last_entry' % url_hash)

        # new entries to be created
        entries = []

        # Dict with fields to grab mapping from rss field name to FlexGet field name
        fields = {
            'guid': 'guid',
            'author': 'author',
            'description': 'description',
            'infohash': 'torrent_info_hash',
        }
        # extend the dict of fields to grab with other_fields list in config
        for field_map in config.get('other_fields', []):
            fields.update(field_map)

        # field name for url can be configured by setting link.
        # default value is auto but for example guid is used in some feeds
        ignored = 0
        for entry in rss.entries:

            # Check if title field is overridden in config
            title_field = config.get('title', 'title')
            # ignore entries without title
            if not entry.get(title_field):
                logger.debug('skipping entry without title')
                ignored += 1
                continue

            # Set the title from the source field
            entry.title = entry[title_field]

            # Check we haven't already processed this entry in a previous run
            if last_entry_id == entry.title + entry.get('guid', ''):
                logger.verbose('Not processing entries from last run.')
                # Let details plugin know that it is ok if this task doesn't produce any entries
                task.no_entries_ok = True
                break

            # remove annoying zero width spaces
            entry.title = entry.title.replace('\u200B', '')

            # helper
            # TODO: confusing? refactor into class member ...

            def add_entry(ea):
                ea['title'] = entry.title

                # fields dict may be modified during this loop, so loop over a copy (fields.items())
                for rss_field, flexget_field in list(fields.items()):
                    if rss_field in entry:
                        if rss_field == 'content':
                            content_str = ''
                            for content in entry[rss_field]:
                                try:
                                    content_str += decode_html(content.value)
                                except UnicodeDecodeError:
                                    logger.warning(
                                        'Failed to decode entry `%s` field `%s`',
                                        ea['title'],
                                        rss_field,
                                    )
                            ea[flexget_field] = content_str
                            logger.debug(
                                'Field `%s` set to `%s` for `%s`',
                                rss_field,
                                ea[rss_field],
                                ea['title'],
                            )
                            continue
                        if not isinstance(getattr(entry, rss_field), str):
                            # Error if this field is not a string
                            logger.error('Cannot grab non text field `{}` from rss.', rss_field)
                            # Remove field from list of fields to avoid repeated error
                            del fields[rss_field]
                            continue
                        if not getattr(entry, rss_field):
                            logger.debug(
                                'Not grabbing blank field %s from rss for %s.',
                                rss_field,
                                ea['title'],
                            )
                            continue
                        try:
                            ea[flexget_field] = decode_html(entry[rss_field])
                            if rss_field in config.get('other_fields', []):
                                # Print a debug message for custom added fields
                                logger.debug(
                                    'Field `%s` set to `%s` for `%s`',
                                    rss_field,
                                    ea[rss_field],
                                    ea['title'],
                                )
                        except UnicodeDecodeError:
                            logger.warning(
                                'Failed to decode entry `%s` field `%s`', ea['title'], rss_field
                            )
                # Also grab pubdate if available
                if hasattr(entry, 'published_parsed') and entry.published_parsed:
                    ea['rss_pubdate'] = datetime(*entry.published_parsed[:6])
                # store basic auth info
                if 'username' in config and 'password' in config:
                    ea['download_auth'] = (config['username'], config['password'])
                entries.append(ea)

            # create from enclosures if present
            enclosures = entry.get('enclosures', [])

            if len(enclosures) > 1 and not config.get('group_links'):
                # There is more than 1 enclosure, create an Entry for each of them
                logger.debug('adding {} entries from enclosures', len(enclosures))
                for enclosure in enclosures:
                    if 'href' not in enclosure:
                        logger.debug('RSS-entry `{}` enclosure does not have URL', entry.title)
                        continue
                    # There is a valid url for this enclosure, create an Entry for it
                    ee = Entry()
                    self.add_enclosure_info(ee, enclosure, config.get('filename', True), True)
                    add_entry(ee)
                # If we created entries for enclosures, we should not create an Entry for the main rss item
                continue

            # create flexget entry
            e = Entry()

            if not isinstance(config.get('link'), list):
                # If the link field is not a list, search for first valid url
                if config['link'] == 'auto':
                    # Auto mode, check for a single enclosure url first
                    if len(entry.get('enclosures', [])) == 1 and entry['enclosures'][0].get(
                        'href'
                    ):
                        self.add_enclosure_info(
                            e, entry['enclosures'][0], config.get('filename', True)
                        )
                    else:
                        # If there is no enclosure url, check link, then guid field for urls
                        for field in ['link', 'guid']:
                            if entry.get(field):
                                e['url'] = entry[field]
                                break
                else:
                    if entry.get(config['link']):
                        e['url'] = entry[config['link']]
            else:
                # If link was passed as a list, we create a list of urls
                for field in config['link']:
                    if entry.get(field):
                        e.setdefault('url', entry[field])
                        if entry[field] not in e.setdefault('urls', []):
                            e['urls'].append(entry[field])

            if config.get('group_links'):
                # Append a list of urls from enclosures to the urls field if group_links is enabled
                enclosure_urls = [enc.href for enc in entry.get('enclosures', [])]
                if enclosure_urls:
                    e.setdefault('url', enclosure_urls[0])
                    e.setdefault('urls', [e['url']])
                    e['urls'].extend(url for url in enclosure_urls if url not in e['urls'])

            if not e.get('url'):
                logger.debug(
                    '{} does not have link ({}) or enclosure', entry.title, config['link']
                )
                ignored += 1
                continue

            add_entry(e)

        # Save last spot in rss
        if rss.entries:
            logger.debug('Saving location in rss feed.')

            try:
                entry_id = rss.entries[0].title + rss.entries[0].get('guid', '')
            except AttributeError:
                entry_id = ''

            if entry_id.strip():
                task.simple_persistence['%s_last_entry' % url_hash] = entry_id
            else:
                logger.debug(
                    'rss feed location saving skipped: no title information in first entry'
                )

        if ignored:
            if not config.get('silent'):
                logger.warning(
                    'Skipped %s RSS-entries without required information (title, link or enclosures)',
                    ignored,
                )

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(InputRSS, 'rss', api_ver=2)



import datetime
from datetime import timedelta

from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, Index, Integer, String, func, select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relation
from sqlalchemy.schema import ForeignKey

from flexget import db_schema
from flexget.event import event
from flexget.utils.database import with_session
from flexget.utils.sqlalchemy_utils import create_index

logger = logger.bind(name='status.db')
Base = db_schema.versioned_base('status', 2)


@db_schema.upgrade('status')
def upgrade(ver, session):
    if ver < 2:
        # Creates the executions table index
        create_index('status_execution', session, 'task_id', 'start', 'end', 'succeeded')
        ver = 2
    return ver


class StatusTask(Base):
    __tablename__ = 'status_task'
    id = Column(Integer, primary_key=True)
    name = Column('task', String)
    executions = relation(
        'TaskExecution', backref='task', cascade='all, delete, delete-orphan', lazy='dynamic'
    )

    def __repr__(self):
        return '<StatusTask(id=%s,name=%s)>' % (self.id, self.name)

    @hybrid_property
    def last_execution_time(self):
        if self.executions.count() == 0:
            return None
        return max(execution.start for execution in self.executions)

    @last_execution_time.expression
    def last_execution_time(cls):
        return (
            select([func.max(TaskExecution.start)])
            .where(TaskExecution.task_id == cls.id)
            .correlate(StatusTask.__table__)
            .label('last_execution_time')
        )

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'last_execution_time': self.last_execution_time
            and self.last_execution_time.astimezone(),
        }


class TaskExecution(Base):
    __tablename__ = 'status_execution'
    id = Column(Integer, primary_key=True)
    task_id = Column(Integer, ForeignKey('status_task.id'))

    start = Column(DateTime)
    end = Column(DateTime)
    succeeded = Column(Boolean, default=True)
    # Entry amounts
    produced = Column(Integer)
    accepted = Column(Integer)
    rejected = Column(Integer)
    failed = Column(Integer)
    abort_reason = Column(String, nullable=True)

    def __repr__(self):
        return '<TaskExecution(task_id=%s,start=%s,end=%s,succeeded=%s,p=%s,a=%s,r=%s,f=%s,reason=%s)>' % (
            self.task_id,
            self.start,
            self.end,
            self.succeeded,
            self.produced,
            self.accepted,
            self.rejected,
            self.failed,
            self.abort_reason,
        )

    def to_dict(self):
        return {
            'id': self.id,
            'task_id': self.task_id,
            'start': self.start and self.start.astimezone(),
            'end': self.end and self.end.astimezone(),
            'succeeded': self.succeeded,
            'produced': self.produced,
            'accepted': self.accepted,
            'rejected': self.rejected,
            'failed': self.failed,
            'abort_reason': self.abort_reason,
        }


Index(
    'ix_status_execution_task_id_start_end_succeeded',
    TaskExecution.task_id,
    TaskExecution.start,
    TaskExecution.end,
    TaskExecution.succeeded,
)


@event('manager.db_cleanup')
def db_cleanup(manager, session):
    # Purge all status data for non existing tasks
    for status_task in session.query(StatusTask).all():
        if status_task.name not in manager.config['tasks']:
            logger.verbose('Purging obsolete status data for task {}', status_task.name)
            session.delete(status_task)

    # Purge task executions older than 1 year
    result = (
        session.query(TaskExecution)
        .filter(TaskExecution.start < datetime.datetime.now() - timedelta(days=365))
        .delete()
    )
    if result:
        logger.verbose('Removed {} task executions from history older than 1 year', result)


@with_session
def get_status_tasks(
    start=None, stop=None, order_by='last_execution_time', descending=True, session=None
):
    logger.debug(
        'querying status tasks: start={}, stop={}, order_by={}, descending={}',
        start,
        stop,
        order_by,
        descending,
    )
    query = session.query(StatusTask)
    if descending:
        query = query.order_by(getattr(StatusTask, order_by).desc())
    else:
        query = query.order_by(getattr(StatusTask, order_by))
    return query.slice(start, stop).all()


@with_session
def get_executions_by_task_id(
    task_id,
    start=None,
    stop=None,
    order_by='start',
    descending=True,
    succeeded=None,
    produced=True,
    start_date=None,
    end_date=None,
    session=None,
):
    logger.debug(
        'querying task executions: task_id={}, start={}, stop={}, order_by={}, '
        'descending={}, succeeded={}, produced={}, start_date={}, end_date={}',
        task_id,
        start,
        stop,
        order_by,
        descending,
        succeeded,
        produced,
        start_date,
        end_date,
    )
    query = session.query(TaskExecution).filter(TaskExecution.task_id == task_id)
    if succeeded:
        query = query.filter(TaskExecution.succeeded == succeeded)
    if produced:
        query = query.filter(TaskExecution.produced > 0)
    if start_date:
        query = query.filter(TaskExecution.start >= start_date)
    if end_date:
        query = query.filter(TaskExecution.start <= end_date)
    if descending:
        query = query.order_by(getattr(TaskExecution, order_by).desc())
    else:
        query = query.order_by(getattr(TaskExecution, order_by))
    return query.slice(start, stop).all()



import os
import re
import time
from functools import total_ordering
from http.client import BadStatusLine
from importlib import import_module
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Union
from urllib.error import HTTPError, URLError

import loguru
import pkg_resources
from requests import RequestException

from flexget import components as components_pkg
from flexget import config_schema
from flexget import plugins as plugins_pkg
from flexget.event import Event
from flexget.event import add_event_handler as add_phase_handler
from flexget.event import event, fire_event, remove_event_handlers

logger = loguru.logger.bind(name='plugin')

PRIORITY_DEFAULT = 128
PRIORITY_LAST = -255
PRIORITY_FIRST = 255


class DependencyError(Exception):
    """Plugin depends on other plugin, but it cannot be loaded.

    Args:
        issued_by: name of the plugin trying to do the import
        missing: name of the plugin or library that is missing
        message: customized user readable error message

    All args are optional.
    """

    def __init__(
        self,
        issued_by: Optional[str] = None,
        missing: Optional[str] = None,
        message: Optional[str] = None,
        silent: bool = False,
    ):
        super().__init__()
        self.issued_by = issued_by
        self.missing = missing
        self._message = message
        self.silent = silent

    def _get_message(self) -> str:
        if self._message:
            return self._message
        return f'Plugin `{self.issued_by}` requires dependency `{self.missing}`'

    def _set_message(self, message: str) -> None:
        self._message = message

    def has_message(self) -> bool:
        return self._message is not None

    message = property(_get_message, _set_message)

    def __str__(self) -> str:
        return '<DependencyError(issued_by=%r,missing=%r,message=%r,silent=%r)>' % (
            self.issued_by,
            self.missing,
            self.message,
            self.silent,
        )


class RegisterException(Exception):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return repr(self.value)


class PluginWarning(Warning):
    def __init__(self, value, logger: 'loguru.Logger' = logger, **kwargs):
        super().__init__()
        self.value = value
        self.logger = logger
        self.kwargs = kwargs

    def __str__(self):
        return self.value


class PluginError(Exception):
    def __init__(self, value, logger: 'loguru.Logger' = logger, **kwargs):
        super().__init__()
        # Value is expected to be a string
        if not isinstance(value, str):
            value = str(value)
        self.value = value
        self.logger = logger
        self.kwargs = kwargs

    def __str__(self):
        return self.value


# TODO: move to utils or somewhere more appropriate
class internet:
    """@internet decorator for plugin phase methods.

    Catches all internet related exceptions and raises PluginError with relevant message.
    Task handles PluginErrors by aborting the task.
    """

    def __init__(self, logger_: 'loguru.Logger' = logger):
        if logger_:
            self.logger = logger_
        else:
            self.logger = logger.bind(name='@internet')

    def __call__(self, func: Callable) -> Callable:
        def wrapped_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except RequestException as e:
                logger.opt(exception=True).debug(
                    'decorator caught RequestException. handled traceback:'
                )
                raise PluginError(f'RequestException: {e}')
            except HTTPError as e:
                raise PluginError(f'HTTPError {e.code}', self.logger)
            except URLError as e:
                logger.opt(exception=True).debug('decorator caught urlerror. handled traceback:')
                raise PluginError(f'URLError {e.reason}', self.logger)
            except BadStatusLine:
                logger.opt(exception=True).debug(
                    'decorator caught badstatusline. handled traceback:'
                )
                raise PluginError('Got BadStatusLine', self.logger)
            except ValueError as e:
                logger.opt(exception=True).debug('decorator caught ValueError. handled traceback:')
                raise PluginError(e)
            except OSError as e:
                logger.opt(exception=True).debug('decorator caught OSError. handled traceback:')
                if hasattr(e, 'reason'):
                    raise PluginError(f'Failed to reach server. Reason: {e.reason}', self.logger)
                if hasattr(e, 'code'):
                    raise PluginError(
                        f"The server couldn't fulfill the request. Error code: {e.code}",
                        self.logger,
                    )
                raise PluginError(f'OSError when connecting to server: {e}', self.logger)

        return wrapped_func


def priority(value: int) -> Callable[[Callable], Callable]:
    """Priority decorator for phase methods"""

    def decorator(target: Callable) -> Callable:
        target.priority = value
        return target

    return decorator


# task phases, in order of their execution; note that this can be extended by
# registering new phases at runtime
task_phases = [
    'prepare',
    'start',
    'input',
    'metainfo',
    'filter',
    'download',
    'modify',
    'output',
    'learn',
    'exit',
]

# map phase names to method names
phase_methods = {
    # task
    'abort': 'on_task_abort'  # special; not a task phase that gets called normally
}
phase_methods.update((_phase, 'on_task_' + _phase) for _phase in task_phases)  # DRY

# Mapping of plugin name to PluginInfo instance (logical singletons)
plugins: Dict[str, 'PluginInfo'] = {}

# Loading done?
plugins_loaded = False

_loaded_plugins = {}
_plugin_options = []
_new_phase_queue: Dict[str, List[Optional[str]]] = {}


def register_task_phase(name: str, before: str = None, after: str = None):
    """Adds a new task phase to the available phases."""
    if before and after:
        raise RegisterException('You can only give either before or after for a phase.')
    if not before and not after:
        raise RegisterException('You must specify either a before or after phase.')
    if name in task_phases or name in _new_phase_queue:
        raise RegisterException(f'Phase {name} already exists.')

    def add_phase(phase_name: str, before: Optional[str], after: Optional[str]):
        if before is not None and before not in task_phases:
            return False
        if after is not None and after not in task_phases:
            return False
        # add method name to phase -> method lookup table
        phase_methods[phase_name] = 'on_task_' + phase_name
        # place phase in phase list
        if before is None and after is not None:
            task_phases.insert(task_phases.index(after) + 1, phase_name)
        if after is None and before is not None:
            task_phases.insert(task_phases.index(before), phase_name)
        return True

    # if can't add yet (dependencies) queue addition
    if not add_phase(name, before, after):
        _new_phase_queue[name] = [before, after]

    for phase_name, args in list(_new_phase_queue.items()):
        if add_phase(phase_name, *args):
            del _new_phase_queue[phase_name]


@total_ordering
class PluginInfo(dict):
    """
    Allows accessing key/value pairs of this dictionary subclass via
    attributes. Also instantiates a plugin and initializes properties.
    """

    # Counts duplicate registrations
    dupe_counter = 0

    def __init__(
        self,
        plugin_class: type,
        name: Optional[str] = None,
        interfaces: Optional[List[str]] = None,
        builtin: bool = False,
        debug: bool = False,
        api_ver: int = 1,
        category: Optional[str] = None,
    ) -> None:
        """
        Register a plugin.

        :param plugin_class: The plugin factory.
        :param string name: Name of the plugin (if not given, default to factory class name in underscore form).
        :param list interfaces: Interfaces this plugin implements.
        :param bool builtin: Auto-activated?
        :param bool debug: True if plugin is for debugging purposes.
        :param int api_ver: Signature of callback hooks (1=task; 2=task,config).
        :param string category: The type of plugin. Can be one of the task phases.
            Defaults to the package name containing the plugin.
        """
        dict.__init__(self)

        if interfaces is None:
            interfaces = ['task']
        if name is None:
            # Convention is to take camel-case class name and rewrite it to an underscore form,
            # e.g. 'PluginName' to 'plugin_name'
            name = re.sub(
                '[A-Z]+', lambda i: '_' + i.group(0).lower(), plugin_class.__name__
            ).lstrip('_')
        if category is None and plugin_class.__module__.startswith('flexget.plugins'):
            # By default look at the containing package of the plugin.
            category = plugin_class.__module__.split('.')[-2]

        # Check for unsupported api versions
        if api_ver < 2:
            raise PluginError('Api versions <2 are no longer supported. Plugin %s' % name)

        # Set basic info attributes
        self.api_ver = api_ver
        self.name = name
        self.interfaces = interfaces
        self.builtin = builtin
        self.debug = debug
        self.category = category
        self.phase_handlers: Dict[str, Event] = {}
        self.schema: config_schema.JsonSchema = {}
        self.schema_id: Optional[str] = None

        self.plugin_class: type = plugin_class
        self.instance: object = None

        if self.name in plugins:
            PluginInfo.dupe_counter += 1
            logger.critical(
                'Error while registering plugin {}. A plugin with the same name is already registered',
                self.name,
            )
        else:
            plugins[self.name] = self

    def initialize(self) -> None:
        if self.instance is not None:
            # We already initialized
            return
        # Create plugin instance
        self.instance = self.plugin_class()
        self.instance.plugin_info = self  # give plugin easy access to its own info
        self.instance.logger = logger.bind(
            name=getattr(self.instance, "LOGGER_NAME", None) or self.name
        )
        if hasattr(self.instance, 'schema'):
            self.schema = self.instance.schema
        elif hasattr(self.instance, 'validator'):
            self.schema = self.instance.validator().schema()
        else:
            # TODO: I think plugins without schemas should not be allowed in config, maybe rethink this
            self.schema = {}

        if self.schema is not None:
            self.schema_id = f'/schema/plugin/{self.name}'
            config_schema.register_schema(self.schema_id, self.schema)

        self.build_phase_handlers()

    def build_phase_handlers(self) -> None:
        """(Re)build phase_handlers in this plugin"""
        for phase, method_name in phase_methods.items():
            if phase in self.phase_handlers:
                continue
            if hasattr(self.instance, method_name):
                method = getattr(self.instance, method_name)
                if not callable(method):
                    continue
                # check for priority decorator
                if hasattr(method, 'priority'):
                    handler_prio = method.priority
                else:
                    handler_prio = PRIORITY_DEFAULT
                event = add_phase_handler(f'plugin.{self.name}.{phase}', method, handler_prio)
                # provides backwards compatibility
                event.plugin = self
                self.phase_handlers[phase] = event

    def __getattr__(self, attr: str):
        if attr in self:
            return self[attr]
        return dict.__getattribute__(self, attr)

    def __setattr__(self, attr: str, value):
        self[attr] = value

    def __str__(self):
        return '<PluginInfo(name=%s)>' % self.name

    def _is_valid_operand(self, other):
        return hasattr(other, 'name')

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    __repr__ = __str__


register = PluginInfo


def _get_standard_plugins_path() -> List[str]:
    """
    :returns: List of directories where traditional plugins should be tried to load from.
    """
    # Get basic path from environment
    paths = []
    env_path = os.environ.get('FLEXGET_PLUGIN_PATH')
    if env_path:
        paths = env_path.split(os.pathsep)

    # Add flexget.plugins directory (core plugins)
    paths.append(os.path.abspath(os.path.dirname(plugins_pkg.__file__)))
    return paths


def _get_standard_components_path() -> List[str]:
    """
    :returns: List of directories where component plugins should be tried to load from.
    """
    # Get basic path from environment
    paths = []
    env_path = os.environ.get('FLEXGET_COMPONENT_PATH')
    if env_path:
        paths = env_path.split(os.pathsep)

    # Add flexget.plugins directory (core plugins)
    paths.append(os.path.abspath(os.path.dirname(components_pkg.__file__)))
    return paths


def _check_phase_queue() -> None:
    if _new_phase_queue:
        for phase, args in _new_phase_queue.items():
            logger.error(
                'Plugin {} requested new phase {}, but it could not be created at requested point (before, after). '
                'Plugin is not working properly.',
                args[0],
                phase,
            )


def _import_plugin(module_name: str, plugin_path: Union[str, Path]) -> None:
    try:
        import_module(module_name)
    except DependencyError as e:
        if e.has_message():
            msg = e.message
        else:
            msg = 'Plugin `%s` requires plugin `%s` to load.' % (
                e.issued_by or module_name,
                e.missing or 'N/A',
            )
        if not e.silent:
            logger.warning(msg)
        else:
            logger.debug(msg)
    except ImportError:
        logger.opt(exception=True).critical(
            'Plugin `{}` failed to import dependencies', module_name
        )
    except ValueError as e:
        # Debugging #2755
        logger.error(
            'ValueError attempting to import `{}` (from {}): {}', module_name, plugin_path, e
        )
    except Exception:
        logger.opt(exception=True).critical('Exception while loading plugin {}', module_name)
        raise
    else:
        logger.trace('Loaded module {} from {}', module_name, plugin_path)


def _load_plugins_from_dirs(dirs: List[str]) -> None:
    """
    :param list dirs: Directories from where plugins are loaded from
    """

    logger.debug(f'Trying to load plugins from: {dirs}')
    dir_paths = [Path(d) for d in dirs if os.path.isdir(d)]
    # add all dirs to plugins_pkg load path so that imports work properly from any of the plugin dirs
    plugins_pkg.__path__ = [str(d) for d in dir_paths]
    for plugins_dir in dir_paths:
        for plugin_path in plugins_dir.glob('**/*.py'):
            if plugin_path.name == '__init__.py':
                continue
            # Split the relative path from the plugins dir to current file's parent dir to find subpackage names
            plugin_subpackages = [
                _f for _f in plugin_path.relative_to(plugins_dir).parent.parts if _f
            ]
            module_name = '.'.join(
                [plugins_pkg.__name__] + plugin_subpackages + [plugin_path.stem]
            )
            _import_plugin(module_name, plugin_path)
    _check_phase_queue()


# TODO: this is now identical to _load_plugins_from_dirs, REMOVE
def _load_components_from_dirs(dirs: List[str]) -> None:
    """
    :param list dirs: Directories where plugin components are loaded from
    """
    logger.debug('Trying to load components from: {}', dirs)
    dir_paths = [Path(d) for d in dirs if os.path.isdir(d)]
    for component_dir in dir_paths:
        for component_path in component_dir.glob('**/*.py'):
            if component_path.name == '__init__.py':
                continue
            # Split the relative path from the plugins dir to current file's parent dir to find subpackage names
            plugin_subpackages = [
                _f for _f in component_path.relative_to(component_dir).parent.parts if _f
            ]
            package_name = '.'.join(
                [components_pkg.__name__] + plugin_subpackages + [component_path.stem]
            )
            _import_plugin(package_name, component_path)
    _check_phase_queue()


def _load_plugins_from_packages() -> None:
    """Load plugins installed via PIP"""
    for entrypoint in pkg_resources.iter_entry_points('FlexGet.plugins'):
        try:
            plugin_module = entrypoint.load()
        except DependencyError as e:
            if e.has_message():
                msg = e.message
            else:
                msg = (
                    'Plugin `%s` requires `%s` to load.',
                    e.issued_by or entrypoint.module_name,
                    e.missing or 'N/A',
                )
            if not e.silent:
                logger.warning(msg)
            else:
                logger.debug(msg)
        except ImportError:
            logger.opt(exception=True).critical(
                'Plugin `{}` failed to import dependencies', entrypoint.module_name
            )
        except Exception:
            logger.opt(exception=True).critical(
                'Exception while loading plugin {}', entrypoint.module_name
            )
            raise
        else:
            logger.trace(
                'Loaded packaged module {} from {}', entrypoint.module_name, plugin_module.__file__
            )
    _check_phase_queue()


def load_plugins(
    extra_plugins: Optional[List[str]] = None, extra_components: Optional[List[str]] = None
) -> None:
    """
    Load plugins from the standard plugin and component paths.

    :param list extra_plugins: Extra directories from where plugins are loaded.
    :param list extra_components: Extra directories from where components are loaded.
    """
    global plugins_loaded

    if extra_plugins is None:
        extra_plugins = []
    if extra_components is None:
        extra_components = []

    # Add flexget.plugins and flexget.components directories (core dist)
    extra_plugins.extend(_get_standard_plugins_path())
    extra_components.extend(_get_standard_components_path())

    start_time = time.time()
    # Import all the plugins
    _load_plugins_from_dirs(extra_plugins)
    _load_components_from_dirs(extra_components)
    _load_plugins_from_packages()
    # Register them
    fire_event('plugin.register')
    # Plugins should only be registered once, remove their handlers after
    remove_event_handlers('plugin.register')
    # After they have all been registered, instantiate them
    for plugin in list(plugins.values()):
        plugin.initialize()
    took = time.time() - start_time
    plugins_loaded = True
    logger.debug(
        'Plugins took {:.2f} seconds to load. {} plugins in registry.', took, len(plugins.keys())
    )


def get_plugins(
    phase: Optional[str] = None,
    interface: Optional[str] = None,
    category: Optional[str] = None,
    name: Optional[str] = None,
    min_api: Optional[int] = None,
) -> Iterable[PluginInfo]:
    """
    Query other plugins characteristics.

    :param string phase: Require phase
    :param string interface: Plugin must implement this interface.
    :param string category: Type of plugin, phase names.
    :param string name: Name of the plugin.
    :param int min_api: Minimum api version.
    :return: List of PluginInfo instances.
    :rtype: list
    """

    def matches(plugin):
        if phase is not None and phase not in phase_methods:
            raise ValueError('Unknown phase %s' % phase)
        if phase and phase not in plugin.phase_handlers:
            return False
        if interface and interface not in plugin.interfaces:
            return False
        if category and not category == plugin.category:
            return False
        if name is not None and name != plugin.name:
            return False
        if min_api is not None and plugin.api_ver < min_api:
            return False
        return True

    return filter(matches, iter(plugins.values()))


def plugin_schemas(**kwargs) -> 'config_schema.JsonSchema':
    """Create a dict schema that matches plugins specified by `kwargs`"""
    return {
        'type': 'object',
        'properties': {p.name: {'$ref': p.schema_id} for p in get_plugins(**kwargs)},
        'additionalProperties': False,
        'error_additionalProperties': '{{message}} Only known plugin names are valid keys.',
        'patternProperties': {'^_': {'title': 'Disabled Plugin'}},
    }


@event('config.register')
def register_schema():
    config_schema.register_schema('/schema/plugins', plugin_schemas)


def get_phases_by_plugin(name: str) -> List[str]:
    """Return all phases plugin :name: hooks"""
    return list(get_plugin_by_name(name).phase_handlers)


def get_plugin_by_name(name: str, issued_by: str = '???') -> PluginInfo:
    """
    Get plugin by name, preferred way since this structure may be changed at some point.

    Getting plugin via `.get` function is recommended for normal use.

    This results much shorter and cleaner code::

        plugin.get_plugin_by_name('parsing').instance.parse_movie(data=entry['title'])

    Shortens into::

        plugin.get('parsing', self).parse_movie(data=entry['title'])

    This function is still useful if you need to access plugin information (PluginInfo).

    :returns PluginInfo instance
    """
    if name not in plugins:
        raise DependencyError(issued_by=issued_by, missing=name)
    return plugins[name]


def get(name: str, requested_by: Union[str, object]) -> object:
    """

    :param str name: Name of the requested plugin
    :param requested_by: Plugin class instance OR string value who is making the request.
    :return: Instance of Plugin class
    """
    if name not in plugins:
        if hasattr(requested_by, 'plugin_info'):
            who = requested_by.plugin_info.name
        else:
            who = requested_by
        raise DependencyError(issued_by=who, missing=name)

    instance = plugins[name].instance
    if instance is None:
        raise Exception('Plugin referred before system initialized?')
    return instance



from loguru import logger

from flexget import entry, plugin
from flexget.components.emby.api_emby import EmbyAuth, EmbyApi
from flexget.components.emby.emby_util import SCHEMA_SERVER, get_field_map
from flexget.event import event

logger = logger.bind(name='emby_lookup')


class EmbyLookup:
    """
    Preforms Emby Lookup

    Example:
        emby_lookup:
            host: http://localhost:8096
            username: <username>
            apikey: <apikey>
            return_host: wan
    """

    auth = {}

    schema = {**SCHEMA_SERVER}

    def on_task_start(self, task, config):
        if not isinstance(config, dict):
            config = {}

        self.auth = EmbyAuth(**config)
        self.auth.login(False)

    @entry.register_lazy_lookup('emby_lookup')
    def lazy_loader(self, entry, auth):
        if not auth:
            return

        if not auth.logged:
            auth.login()

        emby_api = EmbyApi(auth)
        emby_data = emby_api.search(**entry)

        if not emby_data:
            return

        emby_type = EmbyApi.get_type(**emby_data)

        lazy_fields = get_field_map(media_type=emby_type)
        if not lazy_fields:
            return

        entry.update_using_map(lazy_fields, emby_data, ignore_none=True)

    def add_lazy(self, entry, media_type):
        lazy_fields = get_field_map(media_type=media_type)

        entry.add_lazy_fields(self.lazy_loader, lazy_fields, kwargs={'auth': self.auth})

        entry['emby_server_id'] = self.auth.server_id
        entry['emby_username'] = self.auth.username
        entry['emby_user_id'] = self.auth.uid

    # Run after series and metainfo series and imdb
    @plugin.priority(110)
    def on_task_metainfo(self, task, config):
        if not config:
            return

        for entry in task.entries:
            self.add_lazy(entry, EmbyApi.get_type(**entry))

    @property
    def movie_identifier(self):
        """Returns the plugin main identifier type"""
        return 'emby_movie_id'

    @property
    def series_identifier(self):
        """Returns the plugin main identifier type"""
        return 'emby_serie_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        EmbyLookup,
        'emby_lookup',
        api_ver=2,
        interfaces=['task', 'series_metainfo', 'movie_metainfo'],
    )



from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode, func
from sqlalchemy.orm import relationship
from sqlalchemy.sql.elements import and_

from flexget import plugin
from flexget.db_schema import versioned_base, with_session
from flexget.entry import Entry

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.parsing.parsers import parser_common as plugin_parser_common
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='parser_common')

logger = logger.bind(name='movie_list')
Base = versioned_base('movie_list', 0)


class MovieListBase:
    """
    Class that contains helper methods for movie list as well as plugins that use it,
    such as API and CLI.
    """

    @property
    def supported_ids(self):
        # Return a list of supported series identifier as registered via their plugins
        return [
            p.instance.movie_identifier for p in plugin.get_plugins(interface='movie_metainfo')
        ]


class MovieListList(Base):
    __tablename__ = 'movie_list_lists'
    id = Column(Integer, primary_key=True)
    name = Column(Unicode, unique=True)
    added = Column(DateTime, default=datetime.now)
    movies = relationship(
        'MovieListMovie', backref='list', cascade='all, delete, delete-orphan', lazy='dynamic'
    )

    def __repr__(self):
        return '<MovieListList,name={}id={}>'.format(self.name, self.id)

    def to_dict(self):
        return {'id': self.id, 'name': self.name, 'added_on': self.added}


class MovieListMovie(Base):
    __tablename__ = 'movie_list_movies'
    id = Column(Integer, primary_key=True)
    added = Column(DateTime, default=datetime.now)
    title = Column(Unicode)
    year = Column(Integer)
    list_id = Column(Integer, ForeignKey(MovieListList.id), nullable=False)
    ids = relationship('MovieListID', backref='movie', cascade='all, delete, delete-orphan')

    def __repr__(self):
        return '<MovieListMovie title=%s,year=%s,list_id=%d>' % (
            self.title,
            self.year,
            self.list_id,
        )

    def to_entry(self, strip_year=False):
        entry = Entry()
        entry['title'] = entry['movie_name'] = self.title
        entry['url'] = 'mock://localhost/movie_list/%d' % self.id
        entry['added'] = self.added
        if self.year:
            if strip_year is False:
                entry['title'] += ' (%d)' % self.year
            entry['movie_year'] = self.year
        for movie_list_id in self.ids:
            entry[movie_list_id.id_name] = movie_list_id.id_value
        return entry

    def to_dict(self):
        return {
            'id': self.id,
            'added_on': self.added,
            'title': self.title,
            'year': self.year,
            'list_id': self.list_id,
            'movies_list_ids': [movie_list_id.to_dict() for movie_list_id in self.ids],
        }

    @property
    def identifiers(self):
        """ Return a dict of movie identifiers """
        return {identifier.id_name: identifier.id_value for identifier in self.ids}


class MovieListID(Base):
    __tablename__ = 'movie_list_ids'
    id = Column(Integer, primary_key=True)
    added = Column(DateTime, default=datetime.now)
    id_name = Column(Unicode)
    id_value = Column(Unicode)
    movie_id = Column(Integer, ForeignKey(MovieListMovie.id))

    def __repr__(self):
        return '<MovieListID id_name=%s,id_value=%s,movie_id=%d>' % (
            self.id_name,
            self.id_value,
            self.movie_id,
        )

    def to_dict(self):
        return {
            'id': self.id,
            'added_on': self.added,
            'id_name': self.id_name,
            'id_value': self.id_value,
            'movie_id': self.movie_id,
        }


@with_session
def get_movies_by_list_id(
    list_id,
    start=None,
    stop=None,
    order_by='added',
    descending=False,
    movie_ids=None,
    session=None,
):
    query = session.query(MovieListMovie).filter(MovieListMovie.list_id == list_id)
    if movie_ids:
        query = query.filter(MovieListMovie.id.in_(movie_ids))
    if descending:
        query = query.order_by(getattr(MovieListMovie, order_by).desc())
    else:
        query = query.order_by(getattr(MovieListMovie, order_by))
    return query.slice(start, stop).all()


@with_session
def get_movie_lists(name=None, session=None):
    logger.debug('retrieving movie lists')
    query = session.query(MovieListList)
    if name:
        logger.debug('filtering by name {}', name)
        query = query.filter(MovieListList.name.contains(name))
    return query.all()


@with_session
def get_list_by_exact_name(name, session=None):
    logger.debug('returning list with name {}', name)
    return (
        session.query(MovieListList).filter(func.lower(MovieListList.name) == name.lower()).one()
    )


@with_session
def get_list_by_id(list_id, session=None):
    logger.debug('fetching list with id {}', list_id)
    return session.query(MovieListList).filter(MovieListList.id == list_id).one()


@with_session
def get_movie_by_id(list_id, movie_id, session=None):
    logger.debug('fetching movie with id {} from list id {}', movie_id, list_id)
    return (
        session.query(MovieListMovie)
        .filter(and_(MovieListMovie.id == movie_id, MovieListMovie.list_id == list_id))
        .one()
    )


@with_session
def get_movie_by_title_and_year(list_id, title, year=None, session=None):
    movie_list = get_list_by_id(list_id=list_id, session=session)
    if movie_list:
        logger.debug('searching for movie {} in list {}', title, list_id)
        return (
            session.query(MovieListMovie)
            .filter(
                and_(
                    func.lower(MovieListMovie.title) == title.lower(),
                    MovieListMovie.year == year,
                    MovieListMovie.list_id == list_id,
                )
            )
            .one_or_none()
        )


@with_session
def get_movie_identifier(identifier_name, identifier_value, movie_id=None, session=None):
    db_movie_id = (
        session.query(MovieListID)
        .filter(
            and_(
                MovieListID.id_name == identifier_name,
                MovieListID.id_value == identifier_value,
                MovieListID.movie_id == movie_id,
            )
        )
        .first()
    )
    if db_movie_id:
        logger.debug('fetching movie identifier {}: {}', db_movie_id.id_name, db_movie_id.id_value)
        return db_movie_id


@with_session
def get_db_movie_identifiers(identifier_list, movie_id=None, session=None):
    db_movie_ids = []
    for identifier in identifier_list:
        for key, value in identifier.items():
            if key in MovieListBase().supported_ids:
                db_movie_id = get_movie_identifier(
                    identifier_name=key, identifier_value=value, movie_id=movie_id, session=session
                )
                if not db_movie_id:
                    logger.debug('creating movie identifier {}: {}', key, value)
                    db_movie_id = MovieListID(id_name=key, id_value=value, movie_id=movie_id)
                session.merge(db_movie_id)
                db_movie_ids.append(db_movie_id)
    return db_movie_ids



from argparse import SUPPRESS
from time import time

from loguru import logger
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import ClauseElement, Executable, _literal_as_text

from flexget import manager, options
from flexget.event import event

logger = logger.bind(name='explain_sql')


class Explain(Executable, ClauseElement):
    def __init__(self, stmt):
        self.statement = _literal_as_text(stmt)


@compiles(Explain)
def explain(element, compiler, **kw):
    text = 'EXPLAIN QUERY PLAN ' + compiler.process(element.statement)
    return text


class ExplainQuery(Query):
    def __iter__(self):
        logger.info('Query:\n\t{}', str(self).replace('\n', '\n\t'))
        explain = self.session.execute(Explain(self)).fetchall()
        text = '\n\t'.join('|'.join(str(x) for x in line) for line in explain)
        before = time()
        result = Query.__iter__(self)
        logger.info('Query Time: {:0.3f} Explain Query Plan: {}', time() - before, text)
        return result


@event('manager.execute.started')
def register_sql_explain(man, options):
    if options.explain_sql:
        manager.Session.kw['query_cls'] = ExplainQuery


@event('manager.execute.completed')
def deregister_sql_explain(man, options):
    if options.explain_sql:
        manager.Session.kw.pop('query_cls', None)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--explain-sql', action='store_true', dest='explain_sql', default=False, help=SUPPRESS
    )



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='extension')


class ModifyExtension:
    """
    Allows specifying file extension explicitly when all other built-in detection mechanisms fail.

    Example:

    extension: nzb
    """

    schema = {'type': ['string', 'number']}

    def on_task_modify(self, task, config):
        ext = str(config)
        if ext.startswith('.'):
            ext = ext[1:]

        for entry in task.entries:
            logger.debug('`{}` filename is `{}`', entry['title'], entry.get('filename', 'N/A'))
            entry['filename'] = '%s.%s' % (entry.get('filename', entry['title']), ext)
            logger.debug('filename is now `{}`', entry['filename'])


@event('plugin.register')
def register_plugin():
    plugin.register(ModifyExtension, 'extension', api_ver=2)



import copy

from flexget.api.app import base_message
from flexget.components.managed_lists.lists.entry_list.api import ObjectsContainer as OC
from flexget.components.managed_lists.lists.entry_list.db import EntryListEntry, EntryListList
from flexget.entry import Entry
from flexget.manager import Session
from flexget.utils import json


class TestEntryListAPI:
    config = 'tasks: {}'

    def test_entry_list_list(self, api_client, schema_match):
        # No params
        rsp = api_client.get('/entry_list/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_return_lists, data)
        assert not errors

        payload = {'name': 'test_list'}

        # Create list
        rsp = api_client.json_post('/entry_list/', data=json.dumps(payload))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_base_object, data)
        assert not errors

        for field, value in payload.items():
            assert data.get(field) == value

        # Named param
        rsp = api_client.get('/entry_list/?name=test_list')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_return_lists, data)
        assert not errors

        for field, value in payload.items():
            assert data[0].get(field) == value

        # Try to Create list again
        rsp = api_client.json_post('/entry_list/', data=json.dumps(payload))
        assert rsp.status_code == 409, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_entry_list_list_id(self, api_client, schema_match):
        payload = {'name': 'test_list'}

        # Create list
        rsp = api_client.json_post('/entry_list/', data=json.dumps(payload))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_base_object, data)
        assert not errors

        for field, value in payload.items():
            assert data.get(field) == value

        # Get list
        rsp = api_client.get('/entry_list/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_base_object, data)
        assert not errors

        for field, value in payload.items():
            assert data.get(field) == value

        # Delete list
        rsp = api_client.delete('/entry_list/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Try to get list
        rsp = api_client.get('/entry_list/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Try to Delete list
        rsp = api_client.delete('/entry_list/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_entry_list_entries(self, api_client, schema_match):
        # Get non existent list
        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload = {'name': 'test_list'}

        # Create list
        rsp = api_client.json_post('/entry_list/', data=json.dumps(payload))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_base_object, data)
        assert not errors

        for field, value in payload.items():
            assert data.get(field) == value

        entry_data = {'title': 'title', 'original_url': 'http://test.com'}

        # Add entry to list
        rsp = api_client.json_post('/entry_list/1/entries/', data=json.dumps(entry_data))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

        for field, value in entry_data.items():
            assert data.get(field) == value

        # Get entries from list
        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_lists_entries_return_object, data)
        assert not errors

        for field, value in entry_data.items():
            assert data[0].get(field) == value

        # Try to re-add entry to list
        rsp = api_client.json_post('/entry_list/1/entries/', data=json.dumps(entry_data))
        assert rsp.status_code == 409, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

        # Try to post to non existing list
        rsp = api_client.json_post('/entry_list/10/entries/', data=json.dumps(entry_data))
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

    def test_entry_list_entry(self, api_client, schema_match):
        payload = {'name': 'test_list'}

        # Create list
        rsp = api_client.json_post('/entry_list/', data=json.dumps(payload))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_base_object, data)
        assert not errors

        for field, value in payload.items():
            assert data.get(field) == value

        entry_data = {'title': 'title', 'original_url': 'http://test.com'}

        # Add entry to list
        rsp = api_client.json_post('/entry_list/1/entries/', data=json.dumps(entry_data))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

        for field, value in entry_data.items():
            assert data.get(field) == value

        # Get entries from list
        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_lists_entries_return_object, data)
        assert not errors

        for field, value in entry_data.items():
            assert data[0].get(field) == value

        # Get specific entry from list
        rsp = api_client.get('/entry_list/1/entries/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

        for field, value in entry_data.items():
            assert data.get(field) == value

        new_entry_data = {'title': 'title2', 'original_url': 'http://test2.com'}

        # Change specific entry from list
        rsp = api_client.json_put('/entry_list/1/entries/1/', data=json.dumps(new_entry_data))
        assert rsp.status_code == 201, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_list_entry_base_object, data)
        assert not errors

        for field, value in new_entry_data.items():
            assert data.get(field) == value

        # Try to change non-existent entry from list
        rsp = api_client.json_put('/entry_list/1/entries/10/', data=json.dumps(new_entry_data))
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Delete specific entry from list
        rsp = api_client.delete('/entry_list/1/entries/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Get non existent entry from list
        rsp = api_client.get('/entry_list/1/entries/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        # Delete non existent entry from list
        rsp = api_client.delete('/entry_list/1/entries/1/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_entry_list_entries_batch_remove(self, api_client, schema_match):
        payload = {'name': 'test_list'}

        # Create list
        api_client.json_post('/entry_list/', data=json.dumps(payload))

        # Add 3 entries to list
        for i in range(3):
            payload = {'title': f'title {i}', 'original_url': f'http://{i}test.com'}
            rsp = api_client.json_post('/entry_list/1/entries/', data=json.dumps(payload))
            assert rsp.status_code == 201

        # get entries is correct
        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_lists_entries_return_object, data)
        assert not errors
        assert len(data) == 3

        payload = {'ids': [1, 2, 3]}

        rsp = api_client.json_delete('entry_list/1/entries/batch', data=json.dumps(payload))
        assert rsp.status_code == 204

        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.entry_lists_entries_return_object, data)
        assert not errors
        assert not data


class TestEntryListPagination:
    config = 'tasks: {}'

    def test_entry_list_pagination(self, api_client, link_headers):
        base_entry = dict(title='test_title_', original_url='url_')
        number_of_entries = 200

        with Session() as session:
            entry_list = EntryListList(name='test list')
            session.add(entry_list)

            for i in range(number_of_entries):
                entry = copy.deepcopy(base_entry)
                for k, v in entry.items():
                    entry[k] = v + str(i)
                e = Entry(entry)
                entry_list.entries.append(EntryListEntry(e, entry_list.id))

        # Default values
        rsp = api_client.get('/entry_list/1/entries/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        # Change page size
        rsp = api_client.get('/entry_list/1/entries/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 100
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 100

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Get different page
        rsp = api_client.get('/entry_list/1/entries/?page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1

    def test_entry_list_sorting(self, api_client):
        base_entry_1 = dict(title='test_title_1', original_url='url_c')
        base_entry_2 = dict(title='test_title_2', original_url='url_b')
        base_entry_3 = dict(title='test_title_3', original_url='url_a')

        with Session() as session:
            entry_list = EntryListList(name='test list')
            session.add(entry_list)

            e1 = Entry(base_entry_1)
            e2 = Entry(base_entry_2)
            e3 = Entry(base_entry_3)

            entry_list.entries.append(EntryListEntry(e1, entry_list.id))
            entry_list.entries.append(EntryListEntry(e2, entry_list.id))
            entry_list.entries.append(EntryListEntry(e3, entry_list.id))

        # Sort by title
        rsp = api_client.get('/entry_list/1/entries/?sort_by=title')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['title'] == 'test_title_3'

        rsp = api_client.get('/entry_list/1/entries/?sort_by=title&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['title'] == 'test_title_1'

        # Sort by original url
        rsp = api_client.get('/entry_list/1/entries/?sort_by=original_url')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['original_url'] == 'url_c'

        rsp = api_client.get('/entry_list/1/entries/?sort_by=original_url&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['original_url'] == 'url_a'

        # Combine sorting and pagination
        rsp = api_client.get('/entry_list/1/entries/?sort_by=title&per_page=2&page=2')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['title'] == 'test_title_1'



import re


class TitleParser:
    propers = ['proper', 'repack', 'rerip', 'real', 'final']

    specials = ['special', 'bonus', 'extra', 'omake', 'ova']

    editions = ['dc', 'extended', 'uncut', 'remastered', 'unrated', 'theatrical', 'chrono', 'se']

    # TODO: All of the quality related keywords can probably be removed from here, as the quality module handles them
    codecs = ['x264', 'x.264', 'h264', 'h.264', 'XViD']

    # lowercase required
    cutoffs = (
        [
            'limited',
            'xvid',
            'h264',
            'x264',
            'h.264',
            'x.264',
            'screener',
            'unrated',
            '3d',
            'extended',
            'directors',
            'director\'s',
            'multisubs',
            'dubbed',
            'subbed',
            'multi',
        ]
        + specials
        + editions
    )

    remove = ['imax']

    sounds = ['AC3', 'DD5.1', 'DTS']

    @staticmethod
    def re_not_in_word(regexp):
        return r'(?<![^\W_])' + regexp + r'(?![^\W_])'

    @staticmethod
    def strip_spaces(text):
        """Removes all unnecessary duplicate spaces from a text"""
        return ' '.join(text.split())

    @staticmethod
    def remove_words(text, words, not_in_word=False):
        """Clean all given :words: from :text: case insensitively"""
        for word in words:
            text = TitleParser.ireplace(text, word, '', not_in_word=not_in_word)
        # remove duplicate spaces
        text = ' '.join(text.split())
        return text

    @staticmethod
    def ireplace(data, old, new, count=0, not_in_word=False):
        """Case insensitive string replace"""
        old = re.escape(old)
        if not_in_word:
            old = TitleParser.re_not_in_word(old)
        pattern = re.compile(old, re.I)
        return re.sub(pattern, new, data, count)



import re
from unicodedata import normalize

from loguru import logger

from flexget import plugin
from flexget.components.sites.utils import torrent_availability
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.requests import RequestException
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='limetorrents')


def clean_symbols(text):
    """Replaces common symbols with spaces. Also normalize unicode strings in decomposed form."""
    result = text
    if isinstance(result, str):
        result = normalize('NFKD', result)
    result = re.sub(r'[ \(\)\-_\[\]\.]+', ' ', result).lower()

    # Leftovers
    result = re.sub(r"[^a-zA-Z0-9 ]", "", result)
    return result


class Limetorrents:
    """
    Limetorrents search plugin.
    """

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'category': {
                        'type': 'string',
                        'enum': [
                            'all',
                            'anime',
                            'applications',
                            'games',
                            'movies',
                            'music',
                            'tv',
                            'other',
                        ],
                        'default': 'all',
                    },
                    'order_by': {'type': 'string', 'enum': ['date', 'seeds'], 'default': 'date'},
                },
                'additionalProperties': False,
            },
        ]
    }

    base_url = 'https://www.limetorrents.cc/'
    errors = False

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for entries on Limetorrents
        """

        if not isinstance(config, dict):
            config = {'category': config}

        order_by = ''
        if isinstance(config.get('order_by'), str):
            if config['order_by'] != 'date':
                order_by = '{0}/1'.format(config['order_by'])

        category = 'all'
        if isinstance(config.get('category'), str):
            category = '{0}'.format(config['category'])

        entries = set()

        for search_string in entry.get('search_strings', [entry['title']]):
            # No special characters - use dashes instead of %20
            cleaned_search_string = clean_symbols(search_string).replace(' ', '-')

            query = 'search/{0}/{1}/{2}'.format(
                category, cleaned_search_string.encode('utf8'), order_by
            )
            logger.debug(
                'Using search: {}; category: {}; ordering: {}',
                cleaned_search_string,
                category,
                order_by or 'default',
            )
            try:
                page = task.requests.get(self.base_url + query)
                logger.debug('requesting: {}', page.url)
            except RequestException as e:
                logger.error('Limetorrents request failed: {}', e)
                continue

            soup = get_soup(page.content)
            if soup.find('a', attrs={'class': 'csprite_dl14'}) is not None:
                for link in soup.findAll('a', attrs={'class': 'csprite_dl14'}):

                    row = link.find_parent('tr')
                    info_url = str(link.get('href'))

                    # Get the title from the URL as it's complete versus the actual Title text which gets cut off
                    title = str(link.next_sibling.get('href'))
                    title = title[: title.rfind('-torrent')].replace('-', ' ')
                    title = title[1:]

                    data = row.findAll('td', attrs={'class': 'tdnormal'})
                    size = str(data[1].text).replace(',', '')

                    seeds = int(row.find('td', attrs={'class': 'tdseed'}).text.replace(',', ''))
                    leeches = int(row.find('td', attrs={'class': 'tdleech'}).text.replace(',', ''))

                    size = parse_filesize(size)

                    e = Entry()

                    e['url'] = info_url
                    e['title'] = title
                    e['torrent_seeds'] = seeds
                    e['torrent_leeches'] = leeches
                    e['torrent_availability'] = torrent_availability(
                        e['torrent_seeds'], e['torrent_leeches']
                    )
                    e['content_size'] = size

                    entries.add(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(Limetorrents, 'limetorrents', interfaces=['search'], api_ver=2)



from loguru import logger
from requests.packages import urllib3

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='verify_ssl')


class VerifySSLCertificates:
    """
    Plugin that can off SSL certificate verification.

    Example::
      verify_ssl_certificates: no
    """

    schema = {'type': 'boolean'}

    @plugin.priority(253)
    def on_task_start(self, task, config):
        if config is False:
            task.requests.verify = False
            # Disabling verification results in a warning for every HTTPS
            # request:
            # "InsecureRequestWarning: Unverified HTTPS request is being made.
            #  Adding certificate verification is strongly advised. See:
            #  https://urllib3.readthedocs.io/en/latest/security.html"
            # Disable those warnings because the user has explicitly disabled
            # verification and the warning is not beneficial.
            # This change is permanent rather than task scoped, but there won't
            # be any warnings to disable when verification is enabled.
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


@event('plugin.register')
def register_plugin():
    plugin.register(VerifySSLCertificates, 'verify_ssl_certificates', api_ver=2)



import datetime

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'pushover'
logger = logger.bind(name=plugin_name)

PUSHOVER_URL = 'https://api.pushover.net/1/messages.json'

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('pushover.net', '5 seconds'))


class PushoverNotifier:
    """
    Example::

      notify:
        entries:
          via:
            - pushover:
                user_key: <USER_KEY> (can also be a list of userkeys)
                token: <TOKEN>
                [device: <DEVICE_STRING>]
                [priority: <PRIORITY>]
                [url: <URL>]
                [url_title: <URL_TITLE>]
                [sound: <SOUND>]
                [retry]: <RETRY>]
                [expire]: <EXPIRE>]
                [callback]: <CALLBACK>]
                [html]: <HTML>]
    """

    schema = {
        'type': 'object',
        'properties': {
            'user_key': one_or_more({'type': 'string'}),
            'api_key': {'type': 'string', 'default': 'aPwSHwkLcNaavShxktBpgJH4bRWc3m'},
            'device': one_or_more({'type': 'string'}),
            'priority': {
                'oneOf': [{'type': 'number', 'minimum': -2, 'maximum': 2}, {'type': 'string'}]
            },
            'url': {'type': 'string'},
            'url_title': {'type': 'string'},
            'sound': {'type': 'string'},
            'retry': {'type': 'integer', 'minimum': 30},
            'expire': {'type': 'integer', 'maximum': 86400},
            'callback': {'type': 'string'},
            'html': {'type': 'boolean'},
        },
        'required': ['user_key'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Sends a Pushover notification

        :param str title: the message's title
        :param str message: the message to send
        :param dict config: The pushover config
        """
        notification = {
            'token': config.get('api_key'),
            'message': message,
            'title': title,
            'device': config.get('device'),
            'priority': config.get('priority'),
            'url': config.get('url'),
            'url_title': config.get('url_title'),
            'sound': config.get('sound'),
            'retry': config.get('retry'),
            'expire': config.get('expire'),
            'callback': config.get('callback'),
        }

        # HTML parsing mode
        if config.get('html'):
            notification['html'] = 1

        # Support multiple devices
        if isinstance(notification['device'], list):
            notification['device'] = ','.join(notification['device'])

        # Special case, verify certain fields exists if priority is 2
        priority = config.get('priority')
        expire = config.get('expire')
        retry = config.get('retry')
        if priority == 2 and not all([expire, retry]):
            logger.warning(
                'Priority set to 2 but fields "expire" and "retry" are not both present.Lowering priority to 1'
            )
            notification['priority'] = 1

        if not isinstance(config['user_key'], list):
            config['user_key'] = [config['user_key']]

        for user in config['user_key']:
            notification['user'] = user
            try:
                response = requests.post(PUSHOVER_URL, data=notification)
            except RequestException as e:
                if e.response is not None:
                    if e.response.status_code == 429:
                        app_reset = e.response.headers.get('X-Limit-App-Reset')
                        if app_reset:
                            reset_time = datetime.datetime.fromtimestamp(int(app_reset)).strftime(
                                '%Y-%m-%d %H:%M:%S'
                            )
                            error_message = (
                                f'Monthly pushover message limit reached. Next reset: {reset_time}'
                            )

                    else:
                        error_message = e.response.json()['errors'][0]
                else:
                    error_message = str(e)
                raise PluginWarning(error_message)

            reset_time = datetime.datetime.fromtimestamp(
                int(response.headers['X-Limit-App-Reset'])
            ).strftime('%Y-%m-%d %H:%M:%S')
            remaining = response.headers['X-Limit-App-Remaining']
            logger.debug(
                'Pushover notification sent. Notifications remaining until next reset: {}. Next reset at: {}',
                remaining,
                reset_time,
            )


@event('plugin.register')
def register_plugin():
    plugin.register(PushoverNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import copy

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event

logger = logger.bind(name='run_task')


class RunTask:
    schema = {
        'type': 'object',
        'properties': {
            'task': one_or_more({'type': 'string'}),
            'when': one_or_more(
                {
                    'type': 'string',
                    'enum': ['accepted', 'rejected', 'failed', 'no_entries', 'aborted', 'always'],
                }
            ),
        },
        'required': ['task'],
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        tasks = config['task']
        if not isinstance(tasks, list):
            config['task'] = [tasks]

        config.setdefault('when', ['always'])

        when = config['when']
        if when and not isinstance(when, list):
            config['when'] = [when]

        return config

    def on_task_exit(self, task, config):
        config = self.prepare_config(config)
        conditions = [
            task.accepted and 'accepted' in config['when'],
            task.rejected and 'rejected' in config['when'],
            task.failed and 'failed' in config['when'],
            not task.all_entries and 'no_entries' in config['when'],
            'always' in config['when'],
        ]
        if any(conditions):
            self.run_tasks(task, config['task'])

    def on_task_abort(self, task, config):
        config = self.prepare_config(config)
        if 'aborted' in config['when']:
            self.run_tasks(task, config['task'])

    def run_tasks(self, current_task, tasks):
        logger.info('Scheduling tasks {} to run', tasks)
        options = copy.copy(current_task.options)
        options.tasks = tasks
        current_task.manager.execute(options=options)


@event('plugin.register')
def register_plugin():
    plugin.register(RunTask, 'run_task', api_ver=2)



from abc import ABC, abstractmethod, abstractstaticmethod
from collections.abc import MutableSet
from datetime import datetime
import re
from urllib.parse import urlencode

import functools
from requests.exceptions import RequestException, HTTPError
from loguru import logger

from flexget.utils.tools import get_current_flexget_version  # , split_title_year
from flexget.utils import requests
from flexget.plugin import PluginError
from flexget.entry import Entry
from flexget.utils.simple_persistence import SimplePersistence
from flexget.components.emby.emby_util import get_field_map

persist = SimplePersistence('api_emby')

EMBY_ENDPOINT_LOGIN = '/emby/Users/AuthenticateByName'
EMBY_ENDPOINT_SEARCH = '/emby/Users/{userid}/Items'
EMBY_ENDPOINT_PHOTOS = '/emby/Items/{itemid}/Images/Primary'
EMBY_ENDPOINT_DOWNLOAD = '/emby/Items/{itemid}/Download'
EMBY_ENDPOINT_GETUSERS = '/emby/Users'
EMBY_ENDPOINT_USERINFO = '/emby/Users/{userid}'
EMBY_ENDPOINT_SERVERINFO = '/emby/System/Info'
EMBY_ENDPOINT_PARENTS = '/emby/Items/{itemid}/Ancestors'
EMBY_ENDPOINT_LIBRARY = '/emby/Library/MediaFolders'
EMBY_ENDPOINT_FAVORITE = '/emby/Users/{userid}/FavoriteItems/{itemid}'
EMBY_ENDPOINT_ITEMUPD = '/emby/Items/{itemid}'
EMBY_ENDPOINT_WATCHED = '/emby/Users/{userid}/PlayedItems/{itemid}'
EMBY_ENDPOINT_NEW_PLAYLIST = '/emby/Playlists'
EMBY_ENDPOINT_PLAYLIST = '/emby/Playlists/{listid}/Items'
EMBY_ENDPOINT_DELETE_ITEM = '/emby/Items/{itemid}'
EMBY_ENDPOINT_LIBRARY_REFRESH = '/emby/Library/Refresh'

logger = logger.bind(name='api_emby')


class EmbyApiBase(ABC):
    """
    Base Class to all API integratios
    """

    EMBY_PREF = 'emby_'

    @staticmethod
    def merge_field_map(dst: dict, *arg: dict, **kwargs):
        """Merge field maps from clild and parent class"""

        allow_new = kwargs.get('allow_new', False)

        destination = dst.copy()

        for src in arg:
            source = src.copy()

            for key in source:
                if key in destination and isinstance(destination[key], str):
                    destination[key] = [destination[key]]
                elif key not in destination and not allow_new:
                    continue
                elif allow_new:
                    destination[key] = []

                if isinstance(source[key], str):
                    source[key] = [source[key]]

                if source[key][0] and source[key][0].find(EmbyApiBase.EMBY_PREF) < 0:
                    source[key].insert(0, f'{EmbyApiBase.EMBY_PREF}{source[key][0]}')

                for value_source in source[key]:
                    if not value_source in destination[key]:
                        destination[key].append(value_source)

        return destination

    @staticmethod
    def update_using_map(target, field_map: dict, source_item, **kwargs):
        """
        Updates based on field map with source
        """

        allow_new = kwargs.get('allow_new', False)

        my_field_map = field_map.copy()

        func_get = dict.get if isinstance(source_item, dict) else getattr
        for field, val in my_field_map.items():

            values = val
            if not isinstance(values, list):
                values = [val]

            if values[0] and values[0].find(EmbyApiBase.EMBY_PREF) < 0:
                values.insert(0, f'{EmbyApiBase.EMBY_PREF}{values[0]}')

            for value in values:
                if isinstance(value, str):
                    try:
                        val = functools.reduce(func_get, value.split('.'), source_item)
                    except TypeError:
                        continue
                else:
                    val = value(source_item)

                if val is None:
                    continue

                if not hasattr(target, field) and not allow_new:
                    continue

                if isinstance(target, dict):
                    target[field] = val
                else:
                    setattr(target, field, val)

                break


class EmbyAuth(EmbyApiBase):
    """
    Manage API Authorizations
    """

    _last_auth = None

    field_map = {
        'host': 'host',
        'return_host': 'return_host',
        '_apikey': 'apikey',
        '_username': 'username',
        '_password': 'password',
    }

    EMBY_DEF_HOST = 'http://localhost:8096'

    EMBY_CLIENT = 'Flexget'
    EMBY_DEVICE = 'Flexget Plugin'
    EMBY_DEVICE_ID = 'flexget_plugin'
    EMBY_VERSION = get_current_flexget_version()

    host = EMBY_DEF_HOST
    return_host = None

    _userid = ''
    _token = ''
    _serverid = ''
    _logged = False
    _username = None
    _password = None
    _apikey = None
    _wanurl = None
    _lanurl = None
    _can_download = None

    def __init__(self, **kwargs):
        if 'server' in kwargs:
            server = kwargs.get('server')
        else:
            server = kwargs

        EmbyApiBase.update_using_map(self, EmbyAuth.field_map, server)

    def login(self, optional=False):
        """Login user to API"""

        userdata = None

        if not self._apikey:
            userdata = self.check_token_data(persist.get('token_data'))
            if not userdata:
                logger.debug('Login to {} with username {}', self.host, self.username)
                args = {'Username': self._username, 'Pw': self._password}

                login_data = EmbyApi.resquest_emby(EMBY_ENDPOINT_LOGIN, self, 'POST', **args)

                if not login_data and optional:
                    return
                elif not login_data:
                    raise PluginError('Could not login to Emby')

                userdata = login_data.get('User')
                self._token = login_data.get('AccessToken')
        else:
            userdata = self.get_user_by_name(self._username)

        if not userdata and optional:
            return
        elif not userdata:
            raise PluginError('Could not login to Emby')

        self._username = userdata.get('Name')
        self._userid = userdata.get('Id')
        self._serverid = userdata.get('ServerId')

        if 'Policy' in userdata and not self._apikey:
            self._can_download = userdata['Policy'].get('EnableContentDownloading', False)
        elif self._apikey:
            self._can_download = True

        self._logged = True

        serverinfo = EmbyApi.resquest_emby(EMBY_ENDPOINT_SERVERINFO, self, 'GET')
        if serverinfo:
            self._wanurl = serverinfo.get('WanAddress')
            self._lanurl = serverinfo.get('LocalAddress')

        self.save_token_data()
        EmbyAuth._last_auth = self

    def logout(self):
        """Logout user from API"""
        self._token = None
        self._logged = False

    def check_token_data(self, token_data):
        """Checks saved tokens"""
        if not token_data:
            return False

        if (
            'token' not in token_data
            or 'userid' not in token_data
            or token_data.get('username') != self._username
            or token_data.get('host') != self.host
        ):
            return False

        self._userid = token_data.get('userid')
        self._token = token_data.get('token')
        self._logged = True
        endpoint = EMBY_ENDPOINT_USERINFO.format(userid=token_data['userid'])
        response = EmbyApi.resquest_emby(endpoint, self, 'GET')
        if not response:
            self.logout()
            return False

        if self._userid != response.get('Id'):
            self.logout()
            return False

        response['AccessToken'] = token_data['token']

        logger.debug('Restored emby token from database')

        return response

    def save_token_data(self):
        """Saves token data to Data Base"""

        if self._apikey or not self.host or not self.username:
            return

        logger.debug('Saving emby token to database')

        persist['token_data'] = {
            'host': self.host.lower(),
            'username': self._username.lower(),
            'userid': self._userid,
            'serverid': self._serverid,
            'token': self._token,
        }

    @property
    def uid(self) -> str:
        return self._userid

    @property
    def username(self) -> str:
        return self._username

    @property
    def logged(self) -> bool:
        return self._logged

    @property
    def can_download(self) -> bool:
        return self._can_download

    @property
    def token(self) -> str:
        return self._token

    @property
    def server_id(self) -> str:
        return self._serverid

    @property
    def wanurl(self) -> str:
        return self._wanurl

    @property
    def lanurl(self) -> str:
        return self._lanurl

    def add_token_header(self, header: dict) -> dict:
        """Adds data to request header"""
        if not header:
            header = {}

        if self._apikey:
            header['X-Emby-Token'] = self._apikey
            return header

        fields = [
            f'Emby UserId={self._userid}',
            f'Client={self.EMBY_CLIENT}',
            f'Device={self.EMBY_DEVICE}',
            f'DeviceId={self.EMBY_DEVICE_ID}',
            f'Version={self.EMBY_VERSION}',
        ]

        header['X-Emby-Authorization'] = ', '.join(fields)
        header['accept'] = 'application/json'

        if not self.logged:
            return header

        header['X-Emby-Token'] = self.token
        return header

    def get_user_by_name(self, name: str) -> dict:
        """Gets user by username"""

        args = {'IsDisabled': False}
        useres = EmbyApi.resquest_emby(EMBY_ENDPOINT_GETUSERS, self, 'GET', **args)
        if not useres:
            return

        for user in useres:
            if user.get('Name').lower() == name.lower():
                return user

    @staticmethod
    def get_last_auth():
        return EmbyAuth._last_auth


class EmbyApiListBase(EmbyApiBase):
    """Base class to all API Lists"""

    auth = None

    id = None
    _name = None
    types = None
    watched = None
    favorite = None
    sort = None

    allow_create = False

    _internal_items = None

    field_map = {
        'id': ['library_id', 'id', 'Id'],
        '_name': ['library_name', 'list', 'name', 'Name'],
        'types': ['types'],
        'watched': ['watched'],
        'favorite': ['favorite'],
        'sort': ['sort'],
    }

    def __init__(self, **kwargs):
        self.auth = EmbyApi.get_auth(**kwargs)
        EmbyApiBase.update_using_map(self, EmbyApiListBase.field_map, kwargs)

        if self.types and not isinstance(self.types, list):
            self.types = [self.types]

        if isinstance(self.sort, str):
            self.sort = {'field': self.sort, 'order': 'ascending'}

        if isinstance(self.sort, dict):
            self.sort['field'] = self.sort['field'].replace('_', '')

    def set_list_search_args(self, args: dict):
        if self.watched is not None:
            args['IsPlayed'] = self.watched

        if self.favorite is not None:
            args['IsFavorite'] = self.favorite

        if self.sort is not None:
            args['SortBy'] = self.sort['field']
            args['SortOrder'] = self.sort['order']

        if not self.types or len(self.types) == 0:
            args['IncludeItemTypes'] = 'Movie,Episode'
        else:
            args['IncludeItemTypes'] = ','.join(self.types)

    def add(self, entry: Entry):
        """Adds a item to list"""
        item = EmbyApiMedia.cast(auth=self.auth, **entry)
        if not item:
            logger.warning('Not possible to match \'{}\' in emby', item.fullname)
            return

        if self.contains(item):
            logger.warning('\'{}\' already in {}', item.fullname, self.fullname)
            return

        logger.debug('Adding \'{}\' to {}', item.fullname, self.fullname)
        self._add(item)

    @abstractmethod
    def _add(self, item: 'EmbyApiMedia'):
        pass

    def remove(self, entry: Entry):
        """Removes a item from list"""
        item = EmbyApiMedia.cast(auth=self.auth, **entry)
        if not item:
            logger.warning('Not possible to match \'{}\' in emby', item.fullname)
            return

        if not self.contains(item):
            logger.warning('\'{}\' not in {}', item.fullname, self.fullname)
            return

        logger.debug('Removing \'{}\' from {}', item.fullname, self.fullname)
        self._remove(item)

    @abstractmethod
    def _remove(self, item: 'EmbyApiMedia'):
        pass

    @abstractmethod
    def get_items(self):
        pass

    def contains(self, item):
        """Checks if list contains item"""
        return bool(self.get(item))

    def get(self, item) -> None:
        """Get Item from list"""

        if isinstance(item, EmbyApiMedia):
            s_item = item
        else:
            s_item = EmbyApiMedia.cast(auth=self.auth, **item)

        if not s_item:
            return None

        if not self.items_raw:
            try:
                next(self.get_items())
            except StopIteration:
                return None

        result = filter(lambda i: i['Id'] == s_item.id, self.items_raw)
        if len(list(result)) > 0:
            return s_item

        return None

    @property
    def name(self):
        return self._name

    @property
    def fullname(self):
        return self._name

    @property
    def len(self):
        return len(self._internal_items) if self._internal_items else 0

    @property
    def created(self):
        return bool(self.id)

    @property
    def items_raw(self):
        return self._internal_items

    @property
    def immutable(self):
        return

    @abstractstaticmethod
    def is_type(**kwargs):
        pass


class EmbyApiList(EmbyApiBase, MutableSet):
    """Class to interface lists"""

    auth = None
    id = None
    name = None
    _list = None
    _items = None

    field_map = {
        'id': ['library_id', 'id', 'Id'],
        'name': ['library_name', 'list', 'name', 'Name'],
    }

    def __init__(self, **kwargs):
        self.auth = EmbyApi.get_auth(**kwargs)
        EmbyApiBase.update_using_map(self, self.field_map, kwargs)

        self._list = self.get_api_list(**kwargs)
        if not self._list:
            raise PluginError('List \'%s\' does not exist' % self.name)

        self._items = self._list.get_items()

    def __contains__(self, item):
        return self._list.contains

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return self._list.len

    def add(self, item) -> None:
        self._list.add(item)

    def discard(self, item) -> None:
        self._list.remove(item)

    def get(self, item) -> None:
        item_s = self._list.get(item)
        if not item_s:
            return None

        return item_s.to_entry()

    @property
    def immutable(self):
        return self._list.immutable

    @staticmethod
    def get_api_list(**kwargs) -> EmbyApiListBase:
        if EmbyApiWatchedList.is_type(**kwargs):
            logger.debug('List is a watched list')
            return EmbyApiWatchedList(**kwargs)
        elif EmbyApiFavoriteList.is_type(**kwargs):
            logger.debug('List is a favorite list')
            return EmbyApiFavoriteList(**kwargs)
        elif EmbyApiLibrary.is_type(**kwargs):
            logger.debug('List is a library')
            return EmbyApiLibrary(**kwargs)
        elif EmbyApiPlayList.is_type(**kwargs):
            logger.debug('List is a playlist')
            return EmbyApiPlayList(**kwargs)

        if EmbyApiWatchedList.is_type(**kwargs) or EmbyApiWatchedList.allow_create:
            logger.debug('Creating a watched list')
            return EmbyApiWatchedList(**kwargs)
        elif EmbyApiFavoriteList.is_type(**kwargs) or EmbyApiFavoriteList.allow_create:
            logger.debug('Creating a favorite list')
            return EmbyApiFavoriteList(**kwargs)
        elif EmbyApiLibrary.is_type(**kwargs) or EmbyApiLibrary.allow_create:
            logger.debug('Creating a library')
            return EmbyApiLibrary(**kwargs)
        elif EmbyApiPlayList.is_type(**kwargs) or EmbyApiPlayList.allow_create:
            logger.debug('Creating a playlist')
            return EmbyApiPlayList(**kwargs)


class EmbyApiLibrary(EmbyApiListBase):
    """Library List"""

    def __init__(self, **kwargs):
        EmbyApiListBase.__init__(self, **kwargs)

        list_data = EmbyApiLibrary._get_list_data(self.auth, list=self.name)
        if not list_data:
            return

        self.id = list_data['Id']

    def _add(self, item):
        pass

    def _remove(self, item):
        pass

    def get_items(self):
        if not self.created:
            return []

        args = {}

        EmbyApi.set_common_search_arg(args)
        self.set_list_search_args(args)

        args['ParentId'] = self.id

        logger.debug('Search library list with: {}', args)
        endpoint = EMBY_ENDPOINT_SEARCH.format(userid=self.auth.uid)
        items = EmbyApi.resquest_emby(endpoint, self.auth, 'GET', **args)
        self._internal_items = items['Items'] if items else []

        for media in self._internal_items:
            media_obj = EmbyApiMedia.cast(auth=self.auth, **media)
            if media_obj:
                yield media_obj

    @staticmethod
    def library_refresh(auth):
        if not auth:
            auth = EmbyApi.get_auth()

        additem = EmbyApi.resquest_emby(EMBY_ENDPOINT_LIBRARY_REFRESH, auth, 'POST')
        if not additem:
            logger.error('Not possible to refresh emby library')

    @staticmethod
    def _get_list_data(auth, **kwargs):
        args = {}

        list_name = kwargs.get('list')

        EmbyApi.set_common_search_arg(args)

        args['IsHidden'] = False

        logger.debug('Search Library name list with: {}', args)
        search_list_data = EmbyApi.resquest_emby(EMBY_ENDPOINT_LIBRARY, auth, 'GET', **args)
        if not search_list_data or not search_list_data['Items']:
            return

        for search_list in search_list_data['Items']:
            if (
                'Type' not in search_list
                or 'CollectionType' not in search_list
                or 'IsFolder' not in search_list
            ):
                continue

            if search_list['Type'] != 'CollectionFolder':
                continue
            elif (
                search_list['CollectionType'] != 'tvshows'
                and search_list['CollectionType'] != 'movies'
            ):
                continue

            if search_list['Name'].lower() == list_name.lower():
                return search_list

    @staticmethod
    def is_type(**kwargs):
        auth = EmbyApi.get_auth(**kwargs)

        list_name = kwargs.get('list')

        list_data = EmbyApiLibrary._get_list_data(auth, list=list_name)
        if not list_data:
            return False

        return True

    @property
    def fullname(self):
        return f'Library \'{self.name}\''

    @property
    def created(self):
        return bool(self.id)

    @property
    def immutable(self):
        return 'Library is not modifiable'


class EmbyApiWatchedList(EmbyApiListBase):
    """Watched Media List"""

    def __init__(self, **kwargs):
        EmbyApiListBase.__init__(self, **kwargs)

        self._name = 'Watched List'

    def _add(self, item: 'EmbyApiMedia'):
        args = {}
        endpoint = EMBY_ENDPOINT_WATCHED.format(userid=self.auth.uid, itemid=item.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'POST', **args)
        if not additem:
            logger.warning('Not possible to add item \'{}\' to watched list', item.fullname)

    def _remove(self, item: 'EmbyApiMedia'):
        args = {}
        endpoint = EMBY_ENDPOINT_WATCHED.format(userid=self.auth.uid, itemid=item.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'DELETE', **args)
        if not additem:
            logger.warning('Not possible to remove item \'{}\' from watched list', item.fullname)

    def get_items(self):
        args = {}

        EmbyApi.set_common_search_arg(args)
        self.set_list_search_args(args)

        args['IsPlayed'] = True

        logger.debug('Search watched list with: {}', args)

        endpoint = EMBY_ENDPOINT_SEARCH.format(userid=self.auth.uid)
        items = EmbyApi.resquest_emby(endpoint, self.auth, 'GET', **args)
        self._internal_items = items['Items'] if items else []

        for media in self._internal_items:
            media_obj = EmbyApiMedia.cast(auth=self.auth, **media)
            if media_obj:
                yield media_obj

    @staticmethod
    def is_type(**kwargs):
        return kwargs.get('list') == 'watched'

    @property
    def created(self):
        return True


class EmbyApiFavoriteList(EmbyApiListBase):
    """Favorite media list"""

    def __init__(self, **kwargs):
        EmbyApiListBase.__init__(self, **kwargs)

        self._name = 'Favorite List'

    def _add(self, item: 'EmbyApiMedia'):
        args = {}
        endpoint = EMBY_ENDPOINT_FAVORITE.format(userid=self.auth.uid, itemid=item.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'POST', **args)
        if not additem:
            logger.warning('Not possible to add item \'{}\' to favorite list', item.fullname)

    def _remove(self, item: 'EmbyApiMedia'):
        args = {}
        endpoint = EMBY_ENDPOINT_FAVORITE.format(userid=self.auth.uid, itemid=item.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'DELETE', **args)
        if not additem:
            logger.warning('Not possible to remove item \'{}\' from favorite list', item.fullname)

    def get_items(self):
        args = {}

        EmbyApi.set_common_search_arg(args)
        self.set_list_search_args(args)

        args['IsFavorite'] = True

        logger.debug('Search favorite list with: {}', args)
        endpoint = EMBY_ENDPOINT_SEARCH.format(userid=self.auth.uid)
        items = EmbyApi.resquest_emby(endpoint, self.auth, 'GET', **args)
        self._internal_items = items['Items'] if items else []

        for media in self._internal_items:
            media_obj = EmbyApiMedia.cast(auth=self.auth, **media)
            if media_obj:
                yield media_obj

    @staticmethod
    def is_type(**kwargs):
        return kwargs.get('list') == 'favorite'

    @property
    def created(self):
        return True


class EmbyApiPlayList(EmbyApiListBase):
    """Playlist lists"""

    allow_create = True

    playlist_bind = {}

    def __init__(self, **kwargs):
        EmbyApiListBase.__init__(self, **kwargs)

        list_data = EmbyApiPlayList._get_list_data(self.auth, list=self.name)
        if not list_data:
            return

        self.id = list_data['Id']

    def _add(self, item: 'EmbyApiMedia'):
        if not self.created:
            new_list = EmbyApiPlayList.create(item, auth=self.auth, list=self.name)
            self.id = new_list.id
            return

        args = {}
        args['UserId'] = self.auth.uid
        args['Ids'] = item.id

        endpoint = EMBY_ENDPOINT_PLAYLIST.format(listid=self.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'POST', **args)
        if not additem:
            logger.warning(
                'Not possible to add item \'{}\' to Playlist \'{}\'', item.fullname, self.name
            )
            return

    def _remove(self, item):
        if not self.playlist_bind:
            self.fill_items()

        if len(self.playlist_bind) == 1:
            logger.debug('{} is empty', self.fullname)
            self.destroy()
            return

        if item.id not in self.playlist_bind:
            logger.warning('Can\'t find entry of \'{}\' in {}', item.fullname, self.fullname)
            return

        args = {}
        args['EntryIds'] = self.playlist_bind[item.id]

        endpoint = EMBY_ENDPOINT_PLAYLIST.format(listid=self.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'DELETE', **args)
        if not additem:
            logger.warning(
                'Not possible to remove item \'{}\' from Playlist \'{}\'', item.fullname, self.name
            )
            return

        self.playlist_bind.pop(item.id, None)

        if not self.playlist_bind:
            self.destroy()

    def destroy(self):
        logger.debug('Deleting {}', self.fullname)
        endpoint = EMBY_ENDPOINT_DELETE_ITEM.format(itemid=self.id)
        additem = EmbyApi.resquest_emby(endpoint, self.auth, 'DELETE')
        if not additem:
            logger.warning('Not possible to delete {}', self.fullname)
            return
        self.id = None
        self._internal_items = []

    def fill_items(self):
        args = {}

        EmbyApi.set_common_search_arg(args)
        self.set_list_search_args(args)

        args['ParentId'] = self.id

        logger.debug('Search PlayList  with: {}', args)
        endpoint = EMBY_ENDPOINT_SEARCH.format(userid=self.auth.uid)
        items = EmbyApi.resquest_emby(endpoint, self.auth, 'GET', **args)
        self._internal_items = items['Items'] if items else []

        for media in self._internal_items:
            self.playlist_bind[media['Id']] = media['PlaylistItemId']

    def get_items(self):
        if not self.created:
            return []

        self.fill_items()

        for media in self._internal_items:
            media_obj = EmbyApiMedia.cast(auth=self.auth, **media)
            if media_obj:
                yield media_obj

    @staticmethod
    def _get_list_data(auth, **kwargs):
        args = {}

        list_name = kwargs.get('list')

        EmbyApi.set_common_search_arg(args)

        args['SearchTerm'] = list_name
        args['Type'] = 'Playlist'

        logger.debug('Search Playlist Name list with: {}', args)
        search_list_data = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not search_list_data or not search_list_data['Items']:
            return

        for search_list in search_list_data['Items']:
            if search_list['Name'].lower() == list_name.lower():
                return search_list

    @staticmethod
    def is_type(**kwargs):
        auth = EmbyApi.get_auth(**kwargs)

        list_name = kwargs.get('list')

        list_data = EmbyApiPlayList._get_list_data(auth, list=list_name)
        if not list_data:
            return False

        return True

    @staticmethod
    def create(item, **kwargs):
        auth = EmbyApi.get_auth(**kwargs)

        list_name = kwargs.get('list')

        args = {}
        args['Name'] = list_name
        args['Ids'] = item.id
        logger.debug('Creating playlist \'{}\'', list_name)

        items = EmbyApi.resquest_emby(EMBY_ENDPOINT_NEW_PLAYLIST, auth, 'POST', **args)
        if not items:
            logger.warning('Not possible to create playlist \'{}\'', list_name)
            return

        logger.debug('Returning information of new playlist \'{}\'', items)

        new_playlist = EmbyApiPlayList(auth=auth, **items)
        if not new_playlist.created:
            logger.warning('Not possible to create playlist \'{}\'', list_name)
            return

        logger.debug('Created playlist \'{}\' with id {}', new_playlist.name, new_playlist.id)

        return new_playlist

    @property
    def fullname(self):
        return f'Playlist \'{self.name}\''


class EmbyApiMedia(EmbyApiBase):
    """ Basic media """

    TYPE = 'unknown'

    auth = None

    id = None
    name = None
    overview = None
    imdb_id = None
    tmdb_id = None
    tvdb_id = None
    aired_date = None

    source_id = None
    mtype = None
    path = None
    created_date = None
    watched = None
    playcount = None
    favorite = None
    media_sources_raw = None
    format_3d = None
    audio = None
    quality = None
    subtitles = None
    photo_tag = None

    library = None

    field_map = {
        'mtype': ['mtype', 'Type'],
        'id': ['id', 'Id'],
        'name': ['name', 'Name'],
        'path': ['path', 'Path'],
        'overview': ['overview', 'Overview'],
        'imdb_id': ['imdb_id', 'ProviderIds.Imdb'],
        'tmdb_id': ['tmdb_id', 'ProviderIds.Tmdb'],
        'tvdb_id': ['tvdb_id', 'ProviderIds.Tvdb'],
        'created_date': ['created_date', 'DateCreated'],
        'aired_date': ['aired_date', 'PremiereDate'],
        'photo_tag': ['photo_tag', 'ImageTags.Primary'],
        'watched': ['watched', 'UserData.Played'],
        'favorite': ['favorite', 'UserData.IsFavorite'],
        'playcount': ['playcount', 'UserData.PlayCount'],
        'format_3d': ['format_3d', 'Video3DFormat'],
    }

    field_to_dic = {}

    def __init__(self, **kwargs):
        if not kwargs:
            return

        myfield_map = EmbyApiMedia.field_map

        if 'field_map' in kwargs:
            myfield_map = EmbyApiBase.merge_field_map(myfield_map, kwargs.get('field_map'))

        EmbyApiBase.update_using_map(self, EmbyApiMedia.field_map, kwargs)

        if 'auth' in kwargs:
            auth = kwargs.get('auth')
        else:
            auth = EmbyAuth()

        self.auth = auth

        if self.created_date:
            self.created_date = EmbyApi.strtotime(self.created_date)

        if self.aired_date:
            self.aired_date = EmbyApi.strtotime(self.aired_date)

        if self.mtype:
            self.mtype = self.mtype.lower()

        if 'MediaSources' in kwargs:
            self.media_sources_raw = kwargs.get('MediaSources')
            for file in self.media_sources_raw:
                if 'Type' not in file or file.get('Type') != 'Default':
                    continue

                if 'MediaStreams' not in file:
                    continue

                self.source_id = file.get('Id')

                for stream in file.get('MediaStreams'):
                    if stream.get('Type') == 'Video':
                        self.quality = stream.get('DisplayTitle')

                    if stream.get('Type') == 'Subtitle' and stream.get('Language'):
                        if not self.subtitles:
                            self.subtitles = []

                        self.subtitles.append(stream['Language'])

                    if stream.get('Type') == 'Audio' and 'Language' in stream:
                        if not self.audio:
                            self.audio = []

                        self.audio.append(stream['Language'])

        self.library = self.get_libary()

    def get_libary(self) -> 'EmbyApiLibrary':
        args = {'userid': self.auth.uid}

        endpoint = EMBY_ENDPOINT_PARENTS.format(itemid=self.id)
        parents = EmbyApi.resquest_emby(endpoint, self.auth, 'GET', **args)
        if not parents:
            return

        if len(parents) < 2:
            return

        parent = parents[len(parents) - 2]

        library = EmbyApiLibrary(auth=self.auth, **parent)

        return library

    def to_entry(self) -> Entry:
        field_map = get_field_map(**self.to_dict())
        entry = Entry()
        entry.update_using_map(field_map, self.to_dict(), ignore_none=True)

        if self.auth:
            entry['emby_server_id'] = self.auth.server_id
            entry['emby_username'] = self.auth.username
            entry['emby_user_id'] = self.auth.uid

        entry['title'] = self.fullname
        entry['url'] = self.page
        return entry

    def to_dict(self) -> dict:
        if not self:
            return {}

        mtype = self.mtype
        if isinstance(mtype, str):
            mtype = mtype.lower()

        return {
            'id': self.id,
            'name': self.name,
            'fullname': self.fullname,
            'type': mtype,
            'mtype': mtype,
            'media_type': mtype,
            'path': self.path,
            'page': self.page,
            'filename': self.filename,
            'file_extension': self.file_extension,
            'created_date': self.created_date,
            'watched': self.watched,
            'year': self.year,
            'favorite': self.favorite,
            'playcount': self.playcount,
            'media_sources_raw': self.media_sources_raw,
            'format_3d': self.format_3d,
            'audio': self.audio,
            'quality': self.quality,
            'subtitles': self.subtitles,
            'imdb_id': self.imdb_id,
            'tmdb_id': self.tmdb_id,
            'tvdb_id': self.tvdb_id,
            'imdb_url': self.imdb_url,
            'executed': True,
            'source_id': self.source_id,
            'download_url': self.download_url,
            'library_name': self.library_name,
        }

    @staticmethod
    def get_mtype(**kwargs) -> str:
        mtype = None
        if 'mtype' in kwargs:
            mtype = kwargs.get('mtype')
        elif 'Type' in kwargs:
            mtype = kwargs.get('Type')
        elif 'emby_type' in kwargs:
            mtype = kwargs.get('emby_type')

        if mtype:
            mtype = mtype.lower()

        return mtype

    @property
    def fullname(self) -> str:
        return self.name

    @property
    def library_name(self) -> str:
        if not self.library:
            return

        return self.library.name

    @property
    def download_url(self) -> str:
        if not self.auth or not self.auth.token or not self.id:
            return

        if not self.auth.can_download:
            return

        endpoint = EMBY_ENDPOINT_DOWNLOAD.format(itemid=self.id)
        qstr = urlencode({'api_key': self.auth.token, 'mediaSourceId': self.source_id})
        return f'{self.host}{endpoint}?{qstr}'

    @property
    def year(self) -> int:
        if self.aired_date:
            return self.aired_date.year

    @property
    def imdb_url(self) -> str:
        if self.imdb_id:
            return f'http://www.imdb.com/title/{self.imdb_id}'

    @property
    def filename(self) -> str:
        if not self.path:
            return

        filename = re.search('([^\\/\\\\]+)$', self.path)
        if filename:
            return filename.group(1)

    @property
    def file_extension(self) -> str:
        if not self.path:
            return

        ext = re.search('\\.([^\\/\\\\]+)$', self.path)
        if ext:
            return ext.group(1)

    @property
    def host(self) -> str:
        if self.auth.return_host == 'lan' and self.auth.lanurl:
            host = f'{self.auth.lanurl}@'
        elif self.auth.return_host == 'wan' and self.auth.wanurl:
            host = f'{self.auth.wanurl}@'
        else:
            return f'{self.auth.host}'

        return host.replace(':80@', '').replace(':443@', '').replace('@', '')

    @property
    def page(self) -> str:
        if not self.id:
            return None

        qstr = urlencode({'id': self.id, 'serverId': self.auth.server_id})
        return f'{self.host}/web/index.html#!/item?{qstr}'

    @property
    def photo(self) -> str:
        if not self.photo_tag or not self.id:
            return None

        endpoint = EMBY_ENDPOINT_PHOTOS.format(itemid=self.id)

        qstr = urlencode(
            {
                'Tag': self.photo_tag,
                'CropWhitespace': 'true',
                'EnableImageEnhancers': 'false',
                'Format': 'jpg',
            }
        )

        return f'{self.host}{endpoint}?{qstr}'

    @staticmethod
    def cast(**kwargs) -> 'EmbyApiMedia':
        if EmbyApiEpisode.is_type(**kwargs):
            return EmbyApiEpisode(**kwargs)

        if EmbyApiSeason.is_type(**kwargs):
            return EmbyApiSeason(**kwargs)

        if EmbyApiSerie.is_type(**kwargs):
            return EmbyApiSerie(**kwargs)

        if EmbyApiMovie.is_type(**kwargs):
            return EmbyApiMovie(**kwargs)

        return EmbyApiMedia(**kwargs)

    @staticmethod
    def search(**kwargs):
        args = {}

        auth = EmbyApi.get_auth(**kwargs)
        EmbyApi.set_provideres_search_arg(args, **kwargs)
        EmbyApi.set_common_search_arg(args)

        args['SearchTerm'] = kwargs.get('title')

        logger.debug('Search media with: {}', args)
        medias = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not medias or 'Items' not in medias:
            logger.warning('No media found')
            return

        media = medias['Items'][0]

        media_api = EmbyApiMedia(auth=auth, **media)

        if media_api.mtype and media_api.mtype != EmbyApiMedia.TYPE:
            return EmbyApi.search(auth=auth, **media_api.to_dict())
        else:
            logger.debug('Found media \'{}\' in emby server', media_api.fullname)

        return media_api


class EmbyApiSerie(EmbyApiMedia):
    """Series"""

    TYPE = 'series'

    field_map_up = {
        'id': 'serie_id',
        'name': ['serie_name', 'series_name', 'SeriesName'],
        'imdb_id': 'serie_imdb_id',
        'tmdb_id': 'serie_tmdb_id',
        'tvdb_id': 'serie_tvdb_id',
        'page': 'serie_page',
        'aired_date': 'serie_aired_date',
        'overview': 'serie_overview',
        'photo': 'serie_photo',
        'serie_id': ['serie_id', 'SeriesId'],
        'serie_name': ['serie_name', 'series_name', 'SeriesName'],
        'serie_imdb_id': 'serie_imdb_id',
        'serie_tmdb_id': 'serie_tmdb_id',
        'serie_tvdb_id': 'serie_tvdb_id',
    }

    def __init__(self, **kwargs):
        EmbyApiMedia.__init__(self, field_map=self.field_map_up, **kwargs)

    def to_dict(self) -> dict:
        if not self:
            return {}

        return {
            **EmbyApiMedia.to_dict(self),
            **{
                'serie_id': self.serie_id,
                'serie_name': self.serie_name,
                'serie_imdb_id': self.serie_imdb_id,
                'serie_tmdb_id': self.serie_tmdb_id,
                'serie_tvdb_id': self.serie_tvdb_id,
                'serie_aired_date': self.serie_aired_date,
                'serie_year': self.serie_year,
                'serie_overview': self.serie_overview,
                'serie_photo': self.serie_photo,
                'serie_page': self.serie_page,
            },
        }

    @property
    def serie_id(self) -> str:
        return self.id

    @property
    def serie_name(self) -> str:
        return self.name

    @property
    def serie_overview(self) -> str:
        return self.overview

    @property
    def serie_photo(self) -> str:
        return self.photo

    @property
    def serie_page(self) -> str:
        return self.page

    @property
    def serie_imdb_id(self) -> str:
        return self.imdb_id

    @property
    def serie_tmdb_id(self) -> str:
        return self.tmdb_id

    @property
    def serie_tvdb_id(self) -> str:
        return self.tvdb_id

    @property
    def serie_aired_date(self) -> 'datetime':
        return self.aired_date

    @property
    def serie_year(self) -> int:
        return self.year

    @staticmethod
    def search(**kwargs):
        args = {}

        auth = EmbyApi.get_auth(**kwargs)

        parameters = {}
        field_map = EmbyApiBase.merge_field_map(
            EmbyApiMedia.field_map,
            EmbyApiSerie.field_map_up,
            allow_new=True,
        )

        EmbyApi.update_using_map(parameters, field_map, kwargs, allow_new=True)

        if 'serie_id' in parameters:
            args['Ids'] = parameters.get('serie_id')
        elif not EmbyApi.set_provideres_search_arg(args, **kwargs):
            args['SearchTerm'] = parameters.get('serie_name')

        EmbyApi.set_common_search_arg(args)
        args['IncludeItemTypes'] = 'Series'

        logger.debug('Search serie with: {}', args)
        series = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not series or 'Items' not in series:
            logger.warning('No serie found')
            return

        if len(series['Items']) == 1:
            serie = series['Items'][0]
        elif len(series['Items']) > 1:
            logger.error('More than one serie found')
            return
        else:
            logger.warning('No serie found')
            return

        serie_api = EmbyApiSerie(auth=auth, **serie)

        logger.debug('Found serie \'{}\' in emby server', serie_api.fullname)

        return serie_api

    @staticmethod
    def is_type(**kwargs) -> bool:
        mtype = EmbyApiMedia.get_mtype(**kwargs)
        return bool(kwargs.get('series_name') or mtype == EmbyApiSerie.TYPE)


class EmbyApiSeason(EmbyApiMedia):
    """Season"""

    TYPE = 'season'

    season = None
    serie = None

    field_map = {'season': ['season', 'IndexNumber']}

    field_map_up = {
        'id': ['season_id', 'SeasonId'],
        'name': 'season_name',
        'imdb_id': 'season_imdb_id',
        'tmdb_id': 'season_tmdb_id',
        'tvdb_id': 'season_tvdb_id',
        'photo': 'season_photo',
        'page': 'season_page',
        'season_id': ['season_id', 'SeasonId'],
        'season_name': 'season_name',
        'season_imdb_id': 'season_imdb_id',
        'season_tmdb_id': 'season_tmdb_id',
        'season_tvdb_id': 'season_tvdb_id',
        'season': ['season', 'series_season'],
    }

    def __init__(self, **kwargs):
        EmbyApiMedia.__init__(self, field_map=self.field_map_up, **kwargs)

        if not kwargs:
            return

        EmbyApiBase.update_using_map(self, EmbyApiSeason.field_map, kwargs)

        if 'api_serie' not in kwargs:
            self.serie = kwargs.get('api_serie')
        else:
            self.serie = EmbyApiSerie.search(**kwargs) or EmbyApiSerie(auth=self.auth)

    def to_dict(self) -> dict:
        if not self:
            return {}

        return {
            **EmbyApiSerie.to_dict(self.serie),
            **EmbyApiMedia.to_dict(self),
            **{
                'season': self.season,
                'season_id': self.season_id,
                'season_name': self.season_name,
                'season_page': self.season_page,
                'season_photo': self.season_photo,
                'season_imdb_id': self.season_imdb_id,
                'season_tmdb_id': self.season_tmdb_id,
                'season_tvdb_id': self.season_tvdb_id,
            },
        }

    @property
    def fullname(self) -> str:
        if not self.serie:
            return self.name
        elif not self.season:
            return f'{self.serie.fullname} Sxx'

        return f'{self.serie.fullname} S{self.season:02d}'

    @property
    def season_name(self) -> str:
        return self.name

    @property
    def season_id(self) -> str:
        return self.id

    @property
    def serie_overview(self) -> str:
        return self.overview

    @property
    def season_imdb_id(self) -> str:
        return self.imdb_id

    @property
    def season_tmdb_id(self) -> str:
        return self.tmdb_id

    @property
    def season_tvdb_id(self) -> str:
        return self.tvdb_id

    @property
    def season_photo(self) -> str:
        return self.photo

    @property
    def season_page(self) -> str:
        return self.page

    @staticmethod
    def is_type(**kwargs) -> bool:
        mtype = EmbyApiMedia.get_mtype(**kwargs)

        return (
            EmbyApiSerie.is_type(**kwargs)
            and 'series_season' in kwargs
            and not EmbyApiEpisode.is_type(**kwargs)
        ) or mtype == EmbyApiSeason.TYPE

    @staticmethod
    def search(**kwargs) -> 'EmbyApiSeason':
        args = {}

        auth = EmbyApi.get_auth(**kwargs)
        kwargs['auth'] = auth

        parameters = {}
        field_map = EmbyApiBase.merge_field_map(
            EmbyApiMedia.field_map,
            EmbyApiSeason.field_map_up,
            EmbyApiSerie.field_map_up,
            allow_new=True,
        )

        EmbyApi.update_using_map(parameters, field_map, kwargs, allow_new=True)

        # We need to have information regarding the series
        episode_serie = None
        if 'api_serie' in kwargs:
            episode_serie = kwargs.get('api_serie')

        if not episode_serie:
            episode_serie = EmbyApiSerie.search(**kwargs)

        if not episode_serie:
            logger.warning('Not possible to determine season, serie not found')
            return

        if 'season_id' in parameters:
            args['Ids'] = parameters.get('season_id')
        else:
            args['ParentId'] = episode_serie.serie_id

        args['IncludeItemTypes'] = 'Season'

        EmbyApi.set_common_search_arg(args)

        logger.debug('Search season with: {}', args)
        seasons = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not seasons or 'Items' not in seasons:
            logger.warning('No season found')
            return

        seasons = seasons['Items']

        target_season = parameters.get('season')

        if target_season:
            seasons_filter = []
            seasons_filter = list(filter(lambda s: s['IndexNumber'] == target_season, seasons))
            seasons = seasons_filter

        if len(seasons) == 1:
            season = seasons[0]
        elif len(seasons) > 1:
            logger.error(
                'More than one season found for {} {}', episode_serie.fullname, target_season
            )
            return
        else:
            logger.warning('No season found')
            return

        season_api = EmbyApiSeason(auth=auth, **season)

        logger.debug('Found season \'{}\' in emby server', season_api.fullname)

        return season_api


class EmbyApiEpisode(EmbyApiMedia):
    """Episode"""

    TYPE = 'episode'

    field_map = {'episode': ['episode', 'IndexNumber']}

    field_map_up = {
        'id': 'ep_id',
        'name': 'ep_name',
        'imdb_id': 'ep_imdb_id',
        'tmdb_id': 'ep_tmdb_id',
        'tvdb_id': 'ep_tvdb_id',
        'aired_date': 'ep_aired_date',
        'photo': 'ep_photo',
        'page': 'ep_page',
        'ep_id': 'ep_id',
        'ep_name': 'ep_name',
        'ep_imdb_id': 'ep_imdb_id',
        'ep_tmdb_id': 'ep_tmdb_id',
        'ep_tvdb_id': 'ep_tvdb_id',
        'episode': ['episode', 'series_episode'],
    }

    episode = None

    def __init__(self, **kwargs):
        EmbyApiMedia.__init__(self, field_map=self.field_map_up, **kwargs)

        if not kwargs:
            return

        EmbyApiBase.update_using_map(self, EmbyApiEpisode.field_map, kwargs)

        if 'api_serie' in kwargs:
            self.serie = kwargs.get('api_serie')
        else:
            self.serie = EmbyApiSerie.search(**kwargs) or EmbyApiSerie(auth=self.auth)

        if 'api_season' in kwargs:
            self.season = kwargs.get('api_season')
        else:
            self.season = EmbyApiSeason.search(**kwargs) or EmbyApiSeason(auth=self.auth)

    def to_dict(self) -> dict:
        if not self:
            return {}

        return {
            **EmbyApiSeason.to_dict(self.season),
            **EmbyApiSerie.to_dict(self.serie),
            **EmbyApiMedia.to_dict(self),
            **{
                'episode': self.episode,
                'ep_id': self.ep_id,
                'ep_name': self.ep_name,
                'ep_page': self.ep_page,
                'ep_photo': self.ep_photo,
                'ep_imdb_id': self.ep_imdb_id,
                'ep_tmdb_id': self.ep_tmdb_id,
                'ep_tvdb_id': self.ep_tvdb_id,
                'ep_aired_date': self.ep_aired_date,
                'ep_overview': self.ep_overview,
            },
        }

    @property
    def fullname(self) -> str:
        if not self.serie:
            return self.name
        elif not self.season:
            return f'{self.serie.fullname} SxxExx'
        elif not self.episode:
            return f'{self.serie.fullname} S{self.season.season:02d}Exx'

        return f'{self.serie.fullname} S{self.season.season:02d}E{self.episode:02d} {self.name}'

    @property
    def ep_id(self) -> str:
        return self.id

    @property
    def ep_name(self) -> str:
        return self.name

    @property
    def ep_imdb_id(self) -> str:
        return self.imdb_id

    @property
    def ep_tmdb_id(self) -> str:
        return self.tmdb_id

    @property
    def ep_tvdb_id(self) -> str:
        return self.tvdb_id

    @property
    def ep_photo(self) -> str:
        return self.photo

    @property
    def ep_overview(self) -> str:
        return self.overview

    @property
    def ep_page(self) -> str:
        return self.page

    @property
    def ep_aired_date(self) -> 'datetime':
        return self.aired_date

    @staticmethod
    def search(**kwargs) -> 'EmbyApiEpisode':
        episode_serie = None
        episode_season = None

        auth = EmbyApi.get_auth(**kwargs)
        kwargs['auth'] = auth

        args = {}

        parameters = {}
        field_map = EmbyApiBase.merge_field_map(
            EmbyApiMedia.field_map,
            EmbyApiEpisode.field_map_up,
            EmbyApiSeason.field_map_up,
            EmbyApiSerie.field_map_up,
            allow_new=True,
        )

        EmbyApi.update_using_map(parameters, field_map, kwargs, allow_new=True)

        # We need to have information regarding the series
        if 'api_serie' in kwargs:
            episode_serie = kwargs.get('api_serie')

        if not episode_serie:
            episode_serie = EmbyApiSerie.search(**kwargs)

        if not episode_serie:
            logger.warning('Not possible to determine episode, serie not found')
            return

        # We need to have information regarding the season
        if 'api_season' in kwargs:
            episode_season = kwargs.get('api_season')

        if not episode_season:
            episode_season = EmbyApiSeason.search(api_serie=episode_serie, **kwargs)

        if not episode_season:
            logger.warning('Not possible to determine episode, season not found')
            return

        # We need to return all the episodes for that show/season
        if 'ep_id' in parameters:
            args['Ids'] = parameters.get('ep_id')
        elif episode_season and episode_season.season_id:
            args['ParentId'] = episode_season.season_id
        elif episode_serie and episode_serie.serie_id:
            args['ParentId'] = episode_serie.serie_id

        EmbyApi.set_common_search_arg(args)
        args['IncludeItemTypes'] = 'Episode'

        logger.debug('Search episode with: {}', args)
        response = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not response or 'Items' not in response:
            logger.warning('No episode found')
            return

        episodes = response['Items']
        if len(episodes) == 0:
            logger.warning('No episode found')
            return

        target_episode = parameters.get('episode')

        episode = []
        episode = list(filter(lambda e: e['IndexNumber'] == target_episode, episodes))

        if len(episode) > 1:
            logger.error("More than one episode found")
            return
        elif len(episode) == 0:
            logger.warning(
                'Episodes found for {} but none matches {}',
                episode_season.fullname,
                target_episode,
            )
            return

        episode_api = EmbyApiEpisode(
            auth=auth,
            api_serie=episode_serie,
            api_season=episode_season,
            **episode[0],
        )

        logger.debug('Found episode \'{}\' in emby server', episode_api.fullname)

        return episode_api

    @staticmethod
    def is_type(**kwargs) -> bool:
        mtype = EmbyApiMedia.get_mtype(**kwargs)

        return (
            EmbyApiSerie.is_type(**kwargs)
            and 'series_season' in kwargs
            and 'series_episode' in kwargs
        ) or mtype == EmbyApiEpisode.TYPE


class EmbyApiMovie(EmbyApiMedia):
    """Movie"""

    TYPE = 'movie'

    field_map_up = {
        'id': 'movie_id',
        'name': 'movie_name',
        'imdb_id': 'movie_imdb_id',
        'tmdb_id': 'movie_tmdb_id',
        'tvdb_id': 'movie_tvdb_id',
        'aired_date': 'movie_aired_date',
        'year': 'movie_year',
        'photo': 'movie_photo',
        'page': 'movie_page',
        'overview': 'movie_overview',
        'movie_id': 'movie_id',
        'movie_name': 'movie_name',
        'movie_imdb_id': 'movie_imdb_id',
        'movie_tmdb_id': 'movie_tmdb_id',
        'movie_tvdb_id': 'movie_tvdb_id',
        'movie_year': 'movie_year',
    }

    def __init__(self, **kwargs):
        EmbyApiMedia.__init__(self, field_map=EmbyApiMovie.field_map_up, **kwargs)
        self.mtype = EmbyApiMovie.TYPE

        if not kwargs:
            return

    def to_dict(self) -> dict:
        if not self:
            return {}

        return {
            **EmbyApiMedia.to_dict(self),
            **{
                'movie_id': self.movie_id,
                'movie_name': self.movie_name,
                'movie_imdb_id': self.movie_imdb_id,
                'movie_tmdb_id': self.movie_tmdb_id,
                'movie_tvdb_id': self.movie_tvdb_id,
                'movie_aired_date': self.movie_aired_date,
                'movie_year': self.movie_year,
                'movie_photo': self.movie_photo,
                'movie_page': self.movie_page,
                'movie_overview': self.movie_overview,
            },
        }

    @property
    def fullname(self):
        return f'{self.name} ({self.year})'

    @property
    def movie_id(self) -> str:
        return self.id

    @property
    def movie_name(self) -> str:
        return self.name

    @property
    def movie_page(self) -> str:
        return self.page

    @property
    def movie_imdb_id(self) -> str:
        return self.imdb_id

    @property
    def movie_tmdb_id(self) -> str:
        return self.tmdb_id

    @property
    def movie_tvdb_id(self) -> str:
        return self.tvdb_id

    @property
    def movie_overview(self) -> str:
        return self.overview

    @property
    def movie_photo(self) -> str:
        return self.photo

    @property
    def movie_aired_date(self) -> 'datetime':
        return self.aired_date

    @property
    def movie_year(self) -> int:
        return self.year

    @staticmethod
    def search(**kwargs) -> 'EmbyApiMovie':
        args = {}

        auth = EmbyApi.get_auth(**kwargs)
        kwargs['auth'] = auth

        parameters = {}
        field_map = EmbyApiBase.merge_field_map(
            EmbyApiMedia.field_map,
            EmbyApiMovie.field_map_up,
            allow_new=True,
        )

        EmbyApi.update_using_map(parameters, field_map, kwargs, allow_new=True)

        EmbyApi.set_common_search_arg(args)

        if 'movie_id' in parameters:
            args['Ids'] = parameters.get('movie_id')
        else:
            if 'movie_name' in parameters:
                args['SearchTerm'] = parameters.get('movie_name')

            if 'movie_year' in parameters:
                args['Years'] = parameters.get('movie_year')

            EmbyApi.set_provideres_search_arg(args, **kwargs)

        args['IncludeItemTypes'] = 'Movie'

        logger.debug('Search movie with: {}', args)
        movies = EmbyApi.resquest_emby(EMBY_ENDPOINT_SEARCH, auth, 'GET', **args)
        if not movies or 'Items' not in movies:
            logger.warning('No movie found')
            return

        if len(movies['Items']) > 1:
            logger.error("More than one movie found")
            return
        elif len(movies['Items']) == 0:
            logger.warning('No movie found')
            return

        movie = movies['Items'][0]

        movie_api = EmbyApiMovie(auth=auth, **movie)

        logger.debug('Found movie {} in emby server', movie_api.fullname)

        return movie_api

    @staticmethod
    def is_type(**kwargs) -> bool:
        mtype = EmbyApiMedia.get_mtype(**kwargs)
        return bool(kwargs.get('movie_name') or mtype == EmbyApiMovie.TYPE.lower())


class EmbyApi(EmbyApiBase):
    """
    Class to interact with Emby API
    """

    _last_auth = None

    auth = None

    EMBY_EXTRA_FIELDS = [
        'DateCreated',
        'Path',
        'ProviderIds',
        'PremiereDate',
        'MediaSources',
        'Video3DFormat',
        'Overview',
    ]

    def __init__(self, auth: 'EmbyAuth'):
        self.auth = auth
        EmbyApi._last_auth = auth

    @staticmethod
    def set_common_search_arg(args: dict):
        args['Recursive'] = True
        args['Fields'] = ','.join(EmbyApi.EMBY_EXTRA_FIELDS)

    @staticmethod
    def set_provideres_search_arg(args: dict, **kwargs):
        providers = []

        if 'imdb_id' in kwargs:
            providers.append(f'imdb.{kwargs.get("imdb_id")}')
        if 'tmdb_id' in kwargs:
            providers.append(f'tmdb.{kwargs.get("tmdb_id")}')
        if 'tvdb_id' in kwargs:
            providers.append(f'tvdb.{kwargs.get("tvdb_id")}')
        if 'tvrage_id' in kwargs:
            providers.append(f'tvrage.{kwargs.get("tvrage_id")}')

        if len(providers) == 0:
            return

        providers = list(dict.fromkeys(providers))

        args['AnyProviderIdEquals'] = ','.join(providers)

    @staticmethod
    def get_auth(**kwargs) -> EmbyAuth:
        if 'auth' in kwargs:
            return kwargs['auth']
        elif EmbyApi._last_auth:
            return EmbyApi._last_auth
        elif EmbyAuth.get_last_auth():
            return EmbyAuth.get_last_auth()

        return EmbyAuth(**kwargs)

    @staticmethod
    def search(**kwargs):
        search_result = None

        if 'auth' not in kwargs:
            kwargs['auth'] = EmbyApi.get_auth(**kwargs)

        if EmbyApiEpisode.is_type(**kwargs):
            logger.debug("API search episode")
            search_result = EmbyApiEpisode.search(**kwargs)
        elif EmbyApiSeason.is_type(**kwargs):
            logger.debug("API search season")
            search_result = EmbyApiSeason.search(**kwargs)
        elif EmbyApiSerie.is_type(**kwargs):
            logger.debug("API search serie")
            search_result = EmbyApiSerie.search(**kwargs)
        elif EmbyApiMovie.is_type(**kwargs):
            logger.debug("API search movie")
            search_result = EmbyApiMovie.search(**kwargs)
        elif 'executed' not in kwargs:
            logger.debug("API search unknown media")
            search_result = EmbyApiMedia.search(**kwargs)

        if not search_result:
            return None

        if isinstance(search_result, dict):
            return search_result
        elif isinstance(search_result, EmbyApiMedia):
            return search_result.to_dict()
        else:
            return None

    @staticmethod
    def search_list(**kwargs):
        if 'auth' not in kwargs:
            kwargs['auth'] = EmbyApi.get_auth(**kwargs)

        mlist = kwargs.get('list')

        list_object = EmbyApiList.get_api_list(**kwargs)
        if not list_object:
            raise PluginError('List \'%s\' does not exist' % mlist)

        mlist = list_object.get_items()

        for list_obj in mlist:
            yield list_obj

    @staticmethod
    def strtotime(date) -> 'datetime':
        # YYYY-MM-DDTHH:MM:SS.0000000+00:00

        if not date:
            return None
        elif isinstance(date, datetime):
            return date
        elif not isinstance(date, str):
            return None

        # Normalize date
        date_py = re.sub(r'^(.*)\.([0-9]{6})[0-9]*\+([0-9]{2})\:([0-9]{2})$', r'\1.\2+\3\4', date)

        try:
            date = datetime.strptime(date_py, '%Y-%m-%dT%H:%M:%S.%f%z')
        except ValueError:
            date = None

        return date

    @staticmethod
    def get_type(**kwargs) -> bool:
        if EmbyApiEpisode.is_type(**kwargs):
            return EmbyApiEpisode.TYPE
        elif EmbyApiSeason.is_type(**kwargs):
            return EmbyApiSeason.TYPE
        elif EmbyApiSerie.is_type(**kwargs):
            return EmbyApiSerie.TYPE
        elif EmbyApiMovie.is_type(**kwargs):
            return EmbyApiMovie.TYPE

        return EmbyApiMedia.TYPE

    @staticmethod
    def resquest_emby(endpoint: str, auth: 'EmbyAuth', method: str, **kwargs):
        verify_certificates = False

        if not auth:
            auth = EmbyApi.get_auth(**kwargs)
            return

        if not auth.host:
            raise PluginError('No Emby server information')

        if auth:
            endpoint = endpoint.format(userid=auth.uid)

        if not auth:
            auth = EmbyApi.get_auth(**kwargs)

        url = f'{auth.host}{endpoint}'

        request_headers = auth.add_token_header({})

        try:
            if method == 'POST':
                response = requests.post(
                    url,
                    json=kwargs,
                    headers=request_headers,
                    allow_redirects=True,
                    verify=verify_certificates,
                )
            elif method == 'GET':
                response = requests.get(
                    url,
                    params=kwargs,
                    headers=request_headers,
                    allow_redirects=True,
                    verify=verify_certificates,
                )
            elif method == 'DELETE':
                response = requests.request(
                    'DELETE',
                    url,
                    params=kwargs,
                    headers=request_headers,
                    allow_redirects=True,
                    verify=verify_certificates,
                )
        except HTTPError as e:  # Autentication Problem
            if e.response.status_code == 401:
                logger.error('Autentication Error: {}', str(e))
                return False
            else:
                raise PluginError('Could not connect to Emby Server: %s' % str(e)) from e
        except RequestException as e:
            raise PluginError('Could not connect to Emby Server: %s' % str(e)) from e

        if response.status_code == 200 or response.status_code == 204:
            try:
                return response.json()
            except ValueError:
                return {'code': response.status_code}
